"""PM/Rel synchronous machine (SPM, IPM, RM) electrical circuit models

"""
import logging
import warnings
import numpy as np
import numpy.linalg as la
from .utils import iqd, betai1, skin_resistance, dqparident, KTH
import scipy.optimize as so
import scipy.interpolate as ip
from functools import partial

logger = logging.getLogger(__name__)


def parident(workdir, engine, temp, machine,
             magnetizingCurves, magnetMat, condMat,
             **kwargs):
    """return dict of parameters of equivalent circuit for PM machines

    Args:
    workdir: directory for intermediate files
    engine: calculation driver (multiproc, docker, condor)

    temp: list of magnet temperatures in degree Celsius
    machine: dict() with machine parameters
    magnetizingCurves: list of dict() with BH curves (or directory)
    magnetMat: list of dict() with magnet material properties
    condMat: list of dict() with conductor material properties

    optional arguments:
    num_cur_steps: number of current steps (default 5)
    num_beta_steps: number of current steps (default 7 per quadrant)
    speed: rotor speed in 1/s (default 160/p)
    i1_max: maximum current in A rms (default approx 3*i1nom)
    period_frac: fraction of rotating angle (default 6)
    cmd: femag executable
    """
    return dqparident(workdir, engine, temp, machine,
                      magnetizingCurves, magnetMat, condMat,
                      **kwargs)


class PmRelMachine(object):
    """Abstract base class for PmRelMachines

    Args:
        m: number of winding phases
        p: number of pole pairs
        r1: stator winding resistance (in Ohm)
        ls: leakage inductance in H
    """

    def __init__(self, m, p, r1, ls, **kwargs):
        self.p = p
        self.m = m
        self.r1 = r1
        self.ls = ls
        self.io = (1, -1)
        self.fo = 50.0
        self.tcu1 = 20
        self.tmag = 20
        self.zeta1 = 0.2
        self.gam = 0.7
        self.kh = 2
        self.kpfe = 1 # iron loss factor
        self.kfric_b = 1
        self.rotor_mass = 0
        self.kth1 = KTH
        self.bertotti = False
        self.losskeys = ['styoke_hyst', 'stteeth_hyst',
                        'styoke_eddy', 'stteeth_eddy',
                        'rotor_hyst', 'rotor_eddy',
                        'magnet']
        # overwritable function: skin_resistance(r0, w, tcu, kth)
        # Arguments:
        # r0: (float) dc resistance in Ohm at 20°C
        # w: (float) frequency in rad
        # tcu: (float) conductor temperature in °C
        # kth: (float) temperature coefficient (default 3.9 e-3)
        self.skin_resistance = None

        # TODO: need this for speedranges and idq_imax_umax mtpv only
        self.check_extrapolation = True
        for k in kwargs.keys():
            setattr(self, k, kwargs[k])
        try:
            self.tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
        except AttributeError:
            self.tfric = 0

        self.plexp = {'styoke_hyst': 1.0,
                      'stteeth_hyst': 1.0,
                      'styoke_eddy': 2.0,
                      'stteeth_eddy': 2.0,
                      'rotor_hyst': 1.0,
                      'rotor_eddy': 2.0}
        def nolosses(x, y):
            return 0
        self._losses = {k: nolosses for k in tuple(self.losskeys)}

    def pfric(self, n):
        """friction and windage losses"""
        return 2*np.pi*n*self.tfric

    def rstat(self, w):
        """stator resistance
        """
        if self.skin_resistance is not None:
            return self.skin_resistance(self.r1, w, self.tcu1, kth=self.kth1)

        return skin_resistance(self.r1, w, self.tcu1, self.zeta1,
                               self.gam, self.kh)

    def torque_iqd(self, iq, id):
        "torque at q-d-current"
        psid, psiq = self.psi(iq, id)
        tq = self.m*self.p/2*(psid*iq - psiq*id)
        return tq

    def iqd_torque(self, torque, iqd0=0, with_mtpa=True):
        """return minimum d-q-current for torque"""
        if np.abs(torque) < 1e-2:
            return (0, 0)
        if np.isscalar(iqd0):
            i0 = self.io
        else:
            i0 = iqd0
        if with_mtpa:
            res = so.minimize(
                lambda iqd: la.norm(iqd), i0, method='SLSQP',
                constraints=({'type': 'eq',
                              'fun': lambda iqd:
                              self.torque_iqd(*iqd) - torque}))
            if res.success:
                #raise ValueError(f'Torque {torque}, io {i0}: {res.message}')
                return res.x
            def func(i1):
                return torque - self.mtpa(i1)[2]
            i1 = so.fsolve(func, res.x[0])[0]
            return self.mtpa(i1)[:2]


        def tqiq(iq):
            return torque - self.torque_iqd(float(iq), 0)
        iq = so.fsolve(tqiq, (i0[0],))[0]
        return iq, 0, self.torque_iqd(iq, 0)

    def iqd_tmech(self, torque, n, iqd0=0, with_mtpa=True):
        """return minimum d-q-current for shaft torque"""
        if np.abs(torque) < 1e-2:
            return (0, 0)
        if np.isscalar(iqd0):
            i0 = self.io
        else:
            i0 = iqd0

        if with_mtpa:
            res = so.minimize(
                lambda iqd: la.norm(iqd), i0, method='SLSQP',
                constraints=({'type': 'eq',
                              'fun': lambda iqd:
                              self.tmech_iqd(*iqd, n) - torque}))
            if res.success:
                return res.x

            #logger.warning("n: %s, torque %s: %s %s",
            #                   60*n, torque, res.message, i0)
            # try a different approach:
            #raise ValueError(
            #    f'Torque {torque:.1f} speed {60*n:.1f} {res.message}')
            def func(i1):
                return torque - self.mtpa_tmech(i1, n)[2]
            i1 = so.fsolve(func, res.x[0])[0]
            return self.mtpa_tmech(i1, n)[:2]

        def tqiq(iq):
            return torque - self.tmech_iqd(float(iq), 0, n)
        iq = so.fsolve(tqiq, (i0[0],))[0]
        return iq, 0, self.tmech_iqd(iq, 0, n)

    def tloss_iqd(self, iq, id, n):
        """return loss torque of d-q current, iron loss correction factor
        and friction windage losses"""
        if n > 1e-3:
            f1 = self.p*n
            plfe = self.kpfe * (self.iqd_plfe1(iq, id, f1) + self.iqd_plfe2(iq, id, f1))
            pmag = self.iqd_plmag(iq, id, f1)
            return (plfe + pmag + self.pfric(n))/(2*np.pi*n)
        return 0

    def tmech_iqd(self, iq, id, n):
        """return shaft torque of d-q current and speed"""
        tq = self.torque_iqd(iq, id)
        return tq - self.tloss_iqd(iq, id, n)

    def uqd(self, w1, iq, id):
        """return uq, ud of frequency w1 and d-q current"""
        psid, psiq = self.psi(iq, id)
        uqd = (self.r1*iq + w1*(self.ls*id + psid),
               self.r1*id - w1*(self.ls*iq + psiq))
        #logger.debug('beta i1 %s u1 %f', betai1(iq, id), la.norm(uqd))
        return uqd

    def w1_umax(self, u, iq, id):
        """return frequency w1 at given voltage u and id, iq current

        Keyword arguments:
        u -- the maximum voltage (RMS)
        iq, id -- the d-q currents"""
        w10 = np.sqrt(2)*u/la.norm(self.psi(iq, id))
        return so.fsolve(
            lambda w1: la.norm(self.uqd(w1, iq, id))-u*np.sqrt(2),
            w10)[0]

    def w1_imax_umax(self, i1max, u1max):
        """return frequency w1 and torque at voltage u1max and current i1max

        Keyword arguments:
        u1max -- the maximum voltage (Vrms)
        i1max -- the maximum current (Arms)"""
        iq, id, T = self.mtpa(i1max)
        n0 = u1max/np.linalg.norm(self.psi(iq, id))/2/2/np.pi/self.p
        sign = -1 if i1max > 0 else 1
        res = so.minimize(
            lambda n: sign*self.mtpa_tmech(i1max, n)[2],
            n0,
            constraints={
                'type': 'eq',
                'fun': lambda n:
                np.sqrt(2)*u1max - la.norm(
                    self.uqd(2*np.pi*n*self.p,
                             *self.mtpa_tmech(i1max, n)[:2]))})
        return 2*np.pi*res.x[0]*self.p, self.mtpa_tmech(i1max, res.x[0])[2]

    def w1_u(self, u, iq, id):
        """return frequency w1 at given voltage u and id, iq current
        (obsolete, use w1_umax)"""
        return self.w1_umax(u, iq, id)

    def w1max(self, u, iq, id):
        """return max frequency w1 at given voltage u and d-q current
        (obsolete, use w1_umax)"""
        return self.w1_umax(u, iq, id)

    def w2_imax_umax(self, imax, umax, maxtorque=True):
        """return frequency at max current and max voltage"""
        w, info, ier, mesg = so.fsolve(lambda x: np.linalg.norm(
            self.uqd(x, *iqd(-np.pi/2, imax))) - umax*np.sqrt(2),
            np.sqrt(2)*umax/la.norm(self.psi(*self.io)),
            full_output=True)
        if ier == 1:
            return w[0]

        logger.warn("w2_imax_umax ier=%d imax %f", ier, imax)
        raise ValueError("w2_imax_umax {} imax {}".format(mesg, imax))

    def beta_u(self, w1, u, i1):
        "beta at given frequency, voltage and current"
        return so.fsolve(lambda b:
                         la.norm(self.uqd(w1, *(iqd(b, i1))))-u*np.sqrt(2),
                         np.arctan2(self.io[1], self.io[0]))[0]

    def iq_u(self, w1, u, id):
        "iq at given frequency, voltage and id current"
        iq0 = max(self.io[0]/4, id*np.tan(self.betarange[0]))
        return so.fsolve(lambda iq:
                         la.norm(
                             self.uqd(w1, iq, np.array([id])))-u*np.sqrt(2),
                         iq0)[0]

    def iqd_uqd(self, w1, uq, ud):
        "return iq, id current at given frequency, voltage"
        return so.fsolve(lambda iqd:
                         np.array((uq, ud)) - self.uqd(w1, *iqd),
                         (0, self.io[1]))

    def i1_tmech(self, torque, beta, n):
        "return i1 current with given torque and beta"
        i1, info, ier, mesg = so.fsolve(
            lambda i1: self.tmech_iqd(*iqd(beta, i1), n)-torque,
            self.i1range[1]/2,
            full_output=True)
        if ier == 1:
            return i1
        raise ValueError("no solution found for torque {}, beta {} io {}".format(
            torque, beta, self.io))

    def i1_torque(self, torque, beta):
        "return i1 current with given torque and beta"
        i1, info, ier, mesg = so.fsolve(
            lambda i1: self.torque_iqd(*iqd(beta, i1))-torque,
            self.i1range[1]/2,
            full_output=True)
        if ier == 1:
            return i1
        raise ValueError("no solution found for torque {}, beta {} io {}".format(
            torque, beta, self.io))

    def i1_voltage(self, w1, u1, beta):
        "return i1 current with given w1, u1 and beta"
        i1, info, ier, mesg = so.fsolve(
            lambda i1: la.norm(self.uqd(w1, *iqd(beta, i1)))-np.sqrt(2)*u1,
            la.norm(self.io),
            full_output=True)
        if ier == 1:
            return i1
        raise ValueError("{} for w1 {}, u1 {}, beta {}".format(
            mesg, w1, u1, beta))

    def id_torque(self, torque, iq):
        "return d current with given torque and d-current"
        id0 = self.iqd_torque(torque)[1]
        return so.fsolve(
            lambda id: self.torque_iqd(np.array([iq]), id)-torque, id0)[0]

    def iqd_tmech_umax(self, torque, w1, u1max, log=0, with_mtpa=True):
        """return d-q current and shaft torque at stator frequency and max voltage
        with minimal current"""
        n = w1/2/np.pi/self.p
        if with_mtpa:
            i0 = self.iqd_tmech(torque, n)
        else:
            i1 = self.i1_tmech(torque, 0, n)
            i0 = iqd(0, i1)

        if np.linalg.norm(self.uqd(w1, *i0))/np.sqrt(2) > u1max:
            beta, i1 = betai1(*i0)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                i0 = iqd(so.fsolve(lambda b: u1max - np.linalg.norm(
                    self.uqd(w1, *iqd(b, i1)))/np.sqrt(2),
                                   beta)[0], i1)

            res = so.minimize(lambda iqd: np.linalg.norm(iqd), i0, method='SLSQP',
                          constraints=(
                              {'type': 'eq',
                               'fun': lambda iqd:
                               self.tmech_iqd(*iqd, n) - torque},
                              {'type': 'ineq',
                               'fun': lambda iqd:
                               np.sqrt(2)*u1max - la.norm(self.uqd(w1, *iqd))}))
            iq, id = res.x
        else:
            iq, id = i0
        if log:
            try:
                log((iq,id))
            except:
                pass  # logger is not correct
        logger.debug("iqd_tmech_umax w1=%f torque=%f %f iq=%f id=%f u1 u1 %f %f",
                     w1, torque, self.torque_iqd(iq, id), iq, id,
                     u1max, np.linalg.norm(
                         self.uqd(w1, iq, id))/np.sqrt(2))
        return iq, id, self.tmech_iqd(iq, id, n)

    def iqd_torque_umax(self, torque, w1, u1max, log=0, with_mtpa=True):
        """return d-q current and torque at stator frequency and max voltage
        with minimal current"""
        if with_mtpa:
            i0 = self.iqd_torque(torque)
        else:
            i1 = self.i1_torque(torque, 0)
            i0 = iqd(0, i1)

        if np.linalg.norm(self.uqd(w1, *i0))/np.sqrt(2) > u1max:
            beta, i1 = betai1(*i0)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                i0 = iqd(so.fsolve(lambda b: u1max - np.linalg.norm(
                    self.uqd(w1, *iqd(b, i1)))/np.sqrt(2),
                                   beta)[0], i1)

        res = so.minimize(lambda iqd: la.norm(iqd), i0, method='SLSQP',
                          constraints=(
                              {'type': 'eq',
                               'fun': lambda iqd:
                               self.torque_iqd(*iqd) - torque},
                              {'type': 'ineq',
                               'fun': lambda iqd:
                               np.sqrt(2)*u1max - la.norm(self.uqd(w1, *iqd))}))

        if log:
            log(res.x)
        logger.debug("iqd_torque_umax w1=%f torque=%f %f iq=%f id=%f u1 u1 %f %f",
                     w1, torque, self.torque_iqd(*res.x), res.x[0], res.x[1],
                     u1max, np.linalg.norm(
                         self.uqd(w1, *res.x))/np.sqrt(2))
        return res.x[0], res.x[1], self.torque_iqd(*res.x)

    def iqd_pmech_imax_umax(self, n, P, i1max, u1max, with_mtpa, with_tmech, log=0):
        """return d-q current and shaft torque at speed n, P const and max voltage"""
        T = P / n / 2 / np.pi
        w1 = 2*np.pi *n * self.p
        logger.debug("field weakening mode %.2f kW @ %.0f rpm %.1f Nm; "
                     "u1=%.0f V; plfric=%.2f W",
                     P/1000, n*60, T, u1max, self.pfric(n))
        iq, id = self.iqd_torque_imax_umax(T, n, u1max, with_tmech=with_tmech)[:2]

        if with_tmech:
            tcon = {'type': 'eq',
                    'fun': lambda iqd:
                    self.tmech_iqd(*iqd, n) - T}
        else:
            tcon = {'type': 'eq',
                    'fun': lambda iqd:
                    self.torque_iqd(*iqd) - T}

        res = so.minimize(lambda iqd: np.linalg.norm(iqd), (iq, id),
                          method='SLSQP',
                          constraints=[tcon,
                                       {'type': 'ineq',
                                        'fun': lambda iqd:
                                        (u1max * np.sqrt(2)) - np.linalg.norm(
                                            self.uqd(w1, *iqd))}])
        if res.success:
            beta, i1 = betai1(*res.x)
            logger.debug("pconst %s i1 %.2f", res.x, betai1(*res.x)[1])
            if log:
                log(res.x)
            if i1 > abs(i1max):
                return self.iqd_imax_umax(i1max, w1, u1max, T,
                                          with_mtpv=False,
                                          with_tmech=with_tmech)
            if with_tmech:
                return *res.x, self.tmech_iqd(*res.x, n)
            else:
                return *res.x, self.torque_iqd(*res.x)

        if with_tmech:
            iq, id, tq = self.mtpv_tmech(w1, u1max, iqd0=res.x, maxtorque=T>0)
        else:
            iq, id, tq = self.mtpv(w1, u1max, iqd0=res.x, maxtorque=T>0)
        logger.debug("mtpa %s i1 %.2f", res.x, betai1(*res.x)[1])
        if log:
            log((iq, id, tq))
        return iq, id, tq

    def iqd_torque_imax_umax(self, torque, n, u1max, with_tmech=False, log=0):
        """return d-q current and torque at stator frequency w1,
        max voltage  and current"""
        if with_tmech:
            iq, id = self.iqd_tmech(torque, n)
        else:
            iq, id = self.iqd_torque(torque)
        w1 = 2*np.pi*n*self.p
        # Constant torque range
        if np.linalg.norm(self.uqd(w1, iq, id)) <= u1max*np.sqrt(2):
            if log:
                log((iq, id, torque))
            return (iq, id, torque)
        # Field weaking range
        imax = betai1(iq, id)[1]
        iq, id, tq = self.iqd_imax_umax(imax, w1, u1max, torque,
                                        with_mtpv=False, with_tmech=with_tmech)
        if log:
            log((iq, id, tq))
        return iq, id, tq

    def iqd_imax_umax(self, i1max, w1, u1max, torque, with_mtpv=True, with_tmech=True):
        """return d-q current and shaft torque at stator frequency and max voltage
        and max current (for motor operation if maxtorque else generator operation)"""

        if torque > 0:
            # -pi/2 --> 0
            b0, b1 = max(-np.pi/2, self.betarange[0]), 0
            if max(self.betarange) < b1:
                raise ValueError(
                    f"invalid betarange for maxtorque>0: {self.betarange}")
        else:
            # -pi/2 --> -pi
            b0, b1 = -np.pi/2, max(-np.pi, self.betarange[0])
            if min(self.betarange) > b0:
                raise ValueError(
                    f"invalid betarange for maxtorque<0: {self.betarange}")

        deps = 1e-6
        kmax = 100

        def u1norm(b):
            return np.linalg.norm(
                self.uqd(w1, *iqd(b, abs(i1max))))/np.sqrt(2)
        if u1norm(b1) < u1max:
            # must reduce current (torque)
            logger.debug('iqd_imax_umax must reduce torque')
            if with_tmech:
                iq, id, tq = self.iqd_tmech_umax(torque, w1, u1max)
            else:
                iq, id, tq = self.iqd_torque_umax(torque, w1, u1max)
            if not with_mtpv:
                return iq, id, tq
            beta, i1 = betai1(iq, id)
        else:
            for k in range(kmax):
                bx = b0 + (b1-b0)/2
                ux = u1norm(bx)
                if ux > u1max:
                    b1 = bx
                else:
                    b0 = bx
                if abs(b1-b0) < deps:
                    break
            beta, i1 = bx, i1max
            iq, id = iqd(beta, abs(i1))
            du = ux - u1max
            logger.debug("iqd_imax_umax n %f beta %f iter=%d du=%f",
                         w1/2/np.pi/self.p*60, beta/np.pi*180, k, du)
            if abs(du) > 0.1:
                logger.debug('oops? iqd_imax_umax one more torque reduction')
                if with_tmech:
                    iq, id = self.iqd_tmech(torque, w1/2/np.pi/self.p,
                                            iq, id)[:2]
                else:
                    iq, id = self.iqd_torque(torque, w1/2/np.pi/self.p,
                                             iq, id)[:2]
        if with_mtpv:
            try:
                if with_tmech:
                    iq, id, tq = self.mtpv_tmech(w1, u1max,
                                                 iqd0=(iq, id),
                                                 maxtorque=torque>0,
                                                 i1max=i1max)
                else:
                    iq, id, tq = self.mtpv(w1, u1max, iqd0=(iq, id),
                                           maxtorque=torque>0,
                                           i1max=i1max)
                return iq, id, tq
            except ValueError as e:
                logger.warning(e)

        if with_tmech:
            tq = self.tmech_iqd(iq, id, w1/2/np.pi/self.p)
        else:
            tq = self.torque_iqd(iq, id)
        logger.debug("iqd_imax_umax w1=%f torque=%f %f iq=%f id=%f u1 %f %f",
                     w1, torque, tq, iq, id, u1max, np.linalg.norm(
                            self.uqd(w1, iq, id))/np.sqrt(2))
        return iq, id, tq

    def mtpa(self, i1):
        """return iq, id, torque at maximum torque of current i1"""
        sign = -1 if i1 > 0 else 1
        b0 = 0 if i1 > 0 else -np.pi
        bopt, fopt, iter, funcalls, warnflag = so.fmin(
            lambda x: sign*self.torque_iqd(*iqd(x, abs(i1))), b0,
            full_output=True,
            disp=0)
        iq, id = iqd(bopt[0], abs(i1))
        return [iq, id, sign*fopt]

    def mtpa_tmech(self, i1, n):
        """return iq, id, shaft torque at maximum torque of current i1"""
        sign = -1 if i1 > 0 else 1
        b0 = 0 if i1 > 0 else -np.pi
        bopt, fopt, iter, funcalls, warnflag = so.fmin(
            lambda x: sign*self.tmech_iqd(*iqd(x, abs(i1)), n), b0,
            full_output=True,
            disp=0)
        iq, id = iqd(bopt[0], abs(i1))
        return [iq, id, sign*fopt]

    def mtpv(self, w1, u1, iqd0=0, maxtorque=True, i1max=0):
        """return d-q-current, torque for voltage and frequency
        with maximum (maxtorque=True) or minimum torque """
        sign = -1 if maxtorque else 1
        if np.isscalar(iqd0):
            i0 = (-sign*self.i1range[1]/20, -self.i1range[1]/np.sqrt(2))
        else:
            i0 = iqd0
        n = w1/2/np.pi/self.p
        constraints=[{
            'type': 'eq',
            'fun': lambda iqd:
            np.sqrt(2)*u1 - la.norm(self.uqd(w1, *iqd))}]
        if i1max:
            constraints.append({'type': 'ineq',
                                'fun': lambda iqd:
                                 i1max - betai1(*iqd)[1]})
        res = so.minimize(lambda iqd: sign*self.torque_iqd(*iqd), i0,
                          method='SLSQP', constraints=constraints)
        #logger.info("mtpv %s", res)
        if res['success']:
            return res.x[0], res.x[1], sign*res.fun
        raise ValueError(f"mtpv w1={w1} u1={u1} maxtorque={maxtorque} res: {res['message']}")

    def mtpv_tmech(self, w1, u1, iqd0=0, maxtorque=True, i1max=0):
        """return d-q-current, shaft torque for voltage and frequency
        with maximum (maxtorque=True) or minimum torque """
        sign = -1 if maxtorque else 1
        if np.isscalar(iqd0):
            i0 = (-sign*self.i1range[1]/20, -self.i1range[1]/np.sqrt(2))
        else:
            i0 = iqd0
        n = w1/2/np.pi/self.p
        constraints=[{'type': 'eq',
                     'fun': lambda iqd:
                     np.sqrt(2)*u1 - la.norm(
                         self.uqd(w1, *iqd))}]
        if i1max:
            constraints.append({'type': 'ineq',
                                'fun': lambda iqd:
                                 i1max - betai1(*iqd)[1]})
        res = so.minimize(lambda iqd: sign*self.tmech_iqd(*iqd, n), i0,
                          method='SLSQP',
                          constraints=constraints)
        #logger.info("mtpv_torque %s", res)
        if res['success']:
            return res.x[0], res.x[1], sign*res.fun
        #logger.warning("w1=%.1f u1=%.1f maxtorque=%s %s: %s", w1, u1, maxtorque, res.x, res.message)
        raise ValueError(f"mtpv_tmech w1={w1} u1={u1} maxtorque={maxtorque} res: {res['message']}")

    def _set_losspar(self, pfe):
        self.fo = pfe['speed']*self.p
        ef = pfe.get('ef', [2.0, 2.0])
        hf = pfe.get('hf', [1.0, 1.0])
        cf = pfe.get('cf', [1.5, 1.5]) # excess losses
        self.plexp = {'styoke_hyst': hf[0],
                      'stteeth_hyst': hf[0],
                      'styoke_eddy': ef[0],
                      'stteeth_eddy': ef[0],
                      'rotor_hyst': hf[1],
                      'rotor_eddy': ef[1]}
        #                          'magnet'):
        if 'styoke_exc' in pfe:
            self.plexp.update({'styoke_exc': cf[0], 
                               'stteeth_exc':cf[0],
                               'rotor_exc': cf[1]}) 

    def betai1_plcu(self, i1, w1=0):
        return self.m*self.rstat(w1)*i1**2

    def iqd_plcu(self, iq, id, w1=0):
        return self.m*self.rstat(w1)*(iq**2+id**2)/2

    def iqd_plcu1(self, iq, id, w1):
        return self.iqd_plcu(iq, id, w1)

    def iqd_plcu2(self, iq, id):
        return np.zeros(np.asarray(iq).shape)

    def betai1_losses(self, beta, i1, f):
        return np.sum([self.betai1_plfe1(beta, i1, f),
                       self.betai1_plfe2(beta, i1, f),
                       self.betai1_plmag(beta, i1, f),
                       self.betai1_plcu(i1, 2*np.pi*f)], axis=0)

    def iqd_losses(self, iq, id, f):
        return np.sum([self.iqd_plfe1(iq, id, f),
                       self.iqd_plfe2(iq, id, f),
                       self.iqd_plmag(iq, id, f),
                       self.iqd_plcu(iq, id, 2*np.pi*f)], axis=0)

    def speedranges(self, i1max, u1max, speedmax,
                    with_pmconst, with_mtpa, with_mtpv, with_tmech):
        """calculate speed range intervals:
        1. const current MTPA (u < u1max)
        2. const voltage: flux reduction / MTPA and MTPV (if enabled)
        returns list of speed limit for each interval
        calculates with friction and windage losses if with_tmech=True
        """
        if with_tmech:
            w1type, T = self.w1_imax_umax(i1max, u1max)
        else:
            iq, id, T = self.mtpa(i1max)
            w1type = self.w1_umax(u1max, iq, id)
        Pmax = w1type/self.p*T
        w1max = 2*np.pi*speedmax*self.p
        wl, wu = [w1type, min(4*w1type, w1max)]
        if with_mtpv:
            kmax = 6
            self.check_extrapolation = False
        else:
            kmax = 0
            k = kmax
        logger.debug("speedranges w1type %f wu %f", wl, wu)
        for k in range(kmax):
            wx = wl + (wu-wl)/2
            try:
                if with_pmconst:
                    iq, id = self.iqd_pmech_imax_umax(
                        wx/np.pi/2/self.p, Pmax, i1max, u1max,
                        with_mtpa, with_tmech)[:2]
                else:
                    iq, id = self.iqd_imax_umax(i1max, wx, u1max,
                                                T, with_mtpv=False)[:2]
                i1 = betai1(iq, id)[1]
                try:
                    if with_tmech:
                        def voltw1(wx):
                            return np.sqrt(2)*i1 - np.linalg.norm(
                                self.mtpv_tmech(wx, u1max, iqd0=(iq, id),
                                                maxtorque=T > 0)[:2])
                    else:
                        def voltw1(wx):
                            return np.sqrt(2)*i1 - np.linalg.norm(
                                self.mtpv(wx, u1max, iqd0=(iq, id),
                                          maxtorque=T > 0)[:2])
                    w, _, ier, _ = so.fsolve(voltw1, wx, full_output=True)
                    logger.debug("fsolve ier %d T %f w %f w1 %f", ier, T, w, wx)
                    if ier in (1, 4, 5):
                        if abs(w[0]) <= wx:
                            self.check_extrapolation = True
                            return [w/2/np.pi/self.p
                                    for w in (w1type, w[0], w1max)]  # ['MTPA', 'MTPV']
                        if w[0] > wu:
                            wl += (wu-wl)/4
                        else:
                            wl = w[0]
                except ValueError as e:
                    logger.debug("%s: wx %f wl %f wu %f", e, wx, wl, wu)
                    wl = wx
                    pass
            except ValueError as e:
                logger.warning(e)
                wu = wx

            logger.debug("%d: wx %f wl %f wu %f --> %f",
                        k, wx, wl, wu, 100*(wu-wl)/wl)

        self.check_extrapolation = True
        w1max = min(w1max, self.w1_umax(
            u1max, *iqd(-np.pi/2, abs(i1max))))
        return [w/2/np.pi/self.p for w in (w1type, w1max)]  # ['MTPA']


    def operating_point(self, T, n, u1max, with_mtpa=True):
        """
        calculate single operating point.

        return dict with values for
            f1 -- (float) stator frequency in Hz
            u1 -- (float) stator phase voltage in V rms
            i1 -- (float) stator phase current in A rms
            beta -- (float) current angle in rad
            id -- (float) D-current in A peak
            iq -- (float) Q-current in A peak
            ud -- (float) D-voltage in A peak
            uq -- (float) Q-voltgae in A peak
            p1 -- (float) electrical power in W
            pmech -- (float) mechanical power in W
            plcu1 -- (float) stator copper losses in W
            plfe1 -- (float) stator iron losses in W
            plfe2 -- (float) rotor iron losses in W
            plmag -- (float) magnet losses in W
            plfric -- (float) friction losses in W, according to Tfric
            losses -- (float) total losses in W
            cosphi -- (float) power factor for this op
            eta -- (float) efficiency for this op
            T -- (float) torque for this op in Nm, copy of argument T
            Tfric -- (float) friction torque in Nm
            n -- (float) speed for this op in 1/s, copy of argument n
            tcu1/2 -- (float) temperature of statur/rotor in deg. C

        Keyword arguments:
            T -- (float) the output torque at the shaft in Nm
            n -- (float) the speed of the machine in 1/s
            u1max -- (float) the maximum phase voltage in V rms
        """

        r = {}  # result dit

        f1 = n * self.p
        w1 = f1 * 2 * np.pi
        iq, id, tq = self.iqd_tmech_umax(T, w1, u1max, with_mtpa=with_mtpa)
        uq, ud = self.uqd(w1, iq, id)
        u1 = la.norm((ud, uq)) / np.sqrt(2.0)
        i1 = la.norm((id, iq)) / np.sqrt(2.0)
        beta = np.arctan2(id, iq)
        gamma = np.arctan2(ud, uq)
        cosphi = np.cos(beta - gamma)
        pmech = tq * n * 2 * np.pi
        plfe1 = self.iqd_plfe1(iq, id, f1)
        plfe2 = self.iqd_plfe2(iq, id, f1)
        plmag = self.iqd_plmag(iq, id, f1)
        plfe = plfe1 + plfe2 + plmag
        plfric = self.pfric(n)
        plcu = self.betai1_plcu(i1, 2 * np.pi * f1)
        pltotal = plfe + plcu + plfric
        p1 = pmech + pltotal
        if np.abs(pmech) < 1e-12:
            eta = 0  # power to low for eta calculation
        elif p1 > pmech:
            eta = pmech/p1  # motor
        else:
            eta = p1/pmech  # generator
        r['i1'] = float(i1)
        r['u1'] = float(u1)
        r['iq'] = float(iq)
        r['id'] = float(id)
        r['beta'] = float(beta/ np.pi * 180)
        r['uq'] = float(uq)
        r['ud'] = float(ud)
        r['pmech'] = float(pmech)
        r['p1'] = float(p1)
        r['plfe1']= float(plfe1)
        r['plfe2'] = float(plfe2)
        r['plmag'] = float(plmag)
        r['plfric'] = float(plfric)
        r['losses'] = float(pltotal)
        r['T'] = float(tq)
        r['Tfric'] = float(self.tfric)
        r['n'] = float(n)
        r['f1'] = float(f1)
        r['eta'] = eta
        r['cosphi'] = cosphi
        r['t_cu1'] = self.tcu1
        r['t_mag'] = self.tmag

        return r

    def characteristics(self, T, n, u1max, nsamples=60,
                        with_mtpv=True, with_mtpa=True,
                        with_pmconst=True, with_tmech=True):
        """calculate torque speed characteristics.
        return dict with list values of
        id, iq, n, T, ud, uq, u1, i1,
        beta, gamma, phi, cosphi, pmech, n_type

        Keyword arguments:
        T -- the maximum torque or the list of torque values in Nm
        n -- the maximum speed or the list of speed values in 1/s
        u1max -- the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        with_mtpv -- (optional) use mtpv if True (default)
        with_pmconst -- (optional) keep pmech const if True (default)
        with_mtpa -- (optional) use mtpa if True (default) in const speed range, set id=0 if false
        with_tmech -- (optional) use friction and windage losses if True (default)
        """
        r = dict(id=[], iq=[], uq=[], ud=[], u1=[], i1=[], T=[],
                 beta=[], gamma=[], phi=[], cosphi=[], pmech=[], n=[])
        if np.isscalar(T):
            if with_mtpa:
                iq, id = self.iqd_torque(T)
                i1max = betai1(iq, id)[1]
                if T < 0:
                    i1max = -i1max
            else:
                i1max = self.i1_torque(T, 0)
                iq, id = iqd(0, i1max)
            if with_tmech:
                w1, Tf = self.w1_imax_umax(i1max, u1max)
            else:
                iq, id = self.iqd_torque(T)
                Tf = T
                w1 = self.w1_umax(u1max, iq, id)
            n1 = w1/2/np.pi/self.p
            r['n_type'] = n1
            nmax = n
            logger.info("Type speed %.4f n: %.4f nmax %.1f T %.1f i1max %.1f",
                        60*n1, 60*n, 60*nmax, Tf, i1max)

            n1 = min(n1, nmax)
            if n1 < nmax:
                interv = 'MTPA',  # fieldweakening range is always MTPA
                if with_mtpa:
                    speedrange = [0] + self.speedranges(
                        i1max, u1max, nmax,
                        with_pmconst, with_mtpa, with_mtpv, with_tmech)
                    if len(speedrange) > 3:
                        interv = 'MTPA', 'MTPV'
                else:
                    speedrange = [0, n1, nmax]
            else:
                interv = []
                nmax = min(nmax, self.w1_umax(
                    u1max, *iqd(-np.pi/2, abs(i1max))))/2/np.pi/self.p
                speedrange = [0, n1, nmax]

            if speedrange[-1] < speedrange[-2]:
                speedrange = speedrange[:-1]
            logger.info("Speedrange T=%g %s", Tf, speedrange)
            n3 = speedrange[-1]
            nstab = [int(nsamples*(x1-x2)/n3)
                     for x1, x2 in zip(speedrange[1:],
                                       speedrange)]
            logger.info("sample intervals %s", nstab)
            for nx in np.linspace(0, n1, nstab[0]):
                if with_tmech:
                    iq, id = self.iqd_tmech(Tf, nx, (iq, id),
                                            with_mtpa or nx == n1)[:2]
                else:
                    iq, id = self.iqd_torque(Tf, (iq, id),
                                             with_mtpa or nx == n1)[:2]
                r['id'].append(id)
                r['iq'].append(iq)
                r['n'].append(nx)
                r['T'].append(Tf)

            Pmax = 2*np.pi*n1*Tf
            for ns, nu, iv in zip(nstab[1:], speedrange[2:], interv):
                # find id, iq, torque in fieldweakening range
                if ns > 0:
                    dn = (nu - r['n'][-1])/ns
                    logger.info("RANGE %s %d: %f -- %f",
                                iv, ns, r['n'][-1] + dn, nu)
                    try:
                        for nn in np.linspace(r['n'][-1]+dn, nu, ns):
                            w1 = 2*np.pi*nn*self.p
                            logger.debug("fieldweakening: n %g T %g i1max %g w1 %g u1 %g",
                                         nn*60, Tf, i1max, w1, u1max)
                            if iv == 'MTPA':
                                if with_pmconst:
                                    iq, id, tq = self.iqd_pmech_imax_umax(
                                        nn, Pmax, i1max, u1max,
                                        with_mtpa=with_mtpa,
                                        with_tmech=with_tmech)
                                else:
                                    iq, id, tq = self.iqd_imax_umax(
                                        i1max, w1, u1max,
                                        Tf, with_tmech=with_tmech,
                                        with_mtpv=(ns == 1))
                            else:
                                if with_tmech:
                                    iq, id, tq = self.mtpv_tmech(w1, u1max,
                                                                 maxtorque=T > 0)
                                else:
                                    iq, id, tq = self.mtpv(w1, u1max,
                                                           maxtorque=T > 0)
                            if (T > 0 and tq > 0) or (T < 0 and tq < 0):
                                r['id'].append(id)
                                r['iq'].append(iq)
                                r['n'].append(nn)
                                r['T'].append(tq)
                            else:
                                logger.warning("fieldweakening: n %g T %g tq %g i1max %g w1 %g u1 %g",
                                               nn*60, T, tq, i1max, w1, u1max)
                    except ValueError as e:
                        nmax = r['n'][-1]
                        logger.warning("%s: adjusted nmax %f T %f", e, nmax, r['T'][-1])
        else:
            for t, nx in zip(T, n):
                w1 = 2*np.pi*nx*self.p
                iq, id, tq = self.iqd_torque_umax(t, w1, u1max)
                r['id'].append(id)
                r['iq'].append(iq)
                r['T'].append(tq)
                r['n'].append(nx)

        for nx, iq, id in zip(r['n'], r['iq'], r['id']):
            w1 = 2*np.pi*nx*self.p
            uq, ud = self.uqd(w1, iq, id)
            r['uq'].append(uq)
            r['ud'].append(ud)
            r['u1'].append(la.norm((ud, uq))/np.sqrt(2.0))
            r['i1'].append(la.norm((id, iq))/np.sqrt(2.0))
            r['beta'].append(np.arctan2(id, iq)/np.pi*180.)
            r['gamma'].append(np.arctan2(ud, uq)/np.pi*180.)

            r['phi'].append(r['beta'][-1] - r['gamma'][-1])
            r['cosphi'].append(np.cos(r['phi'][-1]/180*np.pi))

        if with_tmech:
            pmech = np.array([2*np.pi*nx*tq for nx, tq in zip(r['n'], r['T'])])
        else:
            pmech = np.array([2*np.pi*nx*(tq-self.tfric) for nx, tq in zip(r['n'], r['T'])])
        f1 = np.array(r['n'])*self.p
        plfe1 = self.iqd_plfe1(np.array(r['iq']), np.array(r['id']), f1)
        plfe2 = self.iqd_plfe2(np.array(r['iq']), np.array(r['id']), f1)
        plmag = self.iqd_plmag(np.array(r['iq']), np.array(r['id']), f1)
        plfe = plfe1 + plfe2 + plmag
        plcu = self.betai1_plcu(np.array(r['i1']), 2*np.pi*f1)
        plfw = self.pfric(2*np.pi*f1)
        pltotal = plfe + plcu + plfw
        r['pmech'] = pmech.tolist()
        r['plfe'] = plfe.tolist()
        r['plcu'] = plcu.tolist()
        r['plfw'] = plfw.tolist()
        r['losses'] = pltotal.tolist()
        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()
        return r

    def i1beta_characteristics(self, n_list, i1_list, beta_list, u1max):
        """calculate i1-beta characteristics"""
        r = dict(id=[], iq=[], uq=[], ud=[], u1=[], i1=[], T=[],
                 beta=[], gamma=[], phi=[], cosphi=[], pmech=[], n=[])
        for n, i1, beta in zip(n_list, i1_list, beta_list):
            w1 = 2*np.pi*n*self.p
            beta = beta/180*np.pi
            iq, id = iqd(beta, i1)
            uq, ud = self.uqd(w1, iq, id)
            u1 = la.norm((ud, uq))/np.sqrt(2)
            if u1 > u1max:
                logger.debug("u1 %s > %s", u1, u1max)
                beta = self.beta_u(w1, u1max, i1)
                logger.debug("beta %s", beta*180/np.pi)
                iq, id = iqd(beta, i1)
                logger.debug("beta %s id, %s iq %s", beta*180/np.pi, id, iq)
                uq, ud = self.uqd(w1, iq, id)
                u1 = la.norm((ud, uq))/np.sqrt(2)
                logger.debug("ud %s uq %s --> u1 %s", ud, uq, u1)

            tq = self.tmech_iqd(iq, id)

            r['id'].append(id)
            r['iq'].append(iq)

            r['uq'].append(uq)
            r['ud'].append(ud)
            r['u1'].append(u1)
            r['i1'].append(la.norm((id, iq))/np.sqrt(2))
            r['T'].append(tq)
            r['beta'].append(np.arctan2(id, iq)/np.pi*180.)
            r['gamma'].append(np.arctan2(ud, uq)/np.pi*180.)

            r['n'].append(n)
            r['phi'].append(r['beta'][-1]-r['gamma'][-1])
            r['cosphi'].append(np.cos(r['phi'][-1]/180*np.pi))
            r['pmech'].append(w1/self.p*r['T'][-1])

        r['losses'] = self.iqd_losses(
            *iqd(np.array(beta_list)/180*np.pi,
                 np.array(i1_list)),
            np.array(n_list)*self.p).tolist()
        return r

    def _inrange(self, iqd):
        i1 = np.linalg.norm(iqd)/np.sqrt(2)
        iqmin, idmin = self.iqdmin(i1)
        iqmax, idmax = self.iqdmax(i1)
        return iqmin <= iqd[0] <= iqmax and idmin <= iqd[1] <= idmax


class PmRelMachineLdq(PmRelMachine):
    """Standard set of PM machine given by i1,beta parameters:
    p number of pole pairs
    m number of phases
    psim flux in Vs (RMS)
    ld d-inductance in
    lq q-inductance in H
    r1 stator resistance
    ls stator leakage inductance in H
    beta angle i1 vs up in degrees
    i1 current in A (RMS)

    optional keyword args:
    psid D-Flux in Vs (RMS)
    psiq Q-Flux in Vs (RMS)
    """

    def __init__(self,  m, p, psim=[], ld=[], lq=[],
                 r1=0, beta=[], i1=[], ls=0, **kwargs):

        super(self.__class__, self).__init__(m, p, r1, ls, **kwargs)
        self.psid = None
        self.betarange = (-np.pi, np.pi)
        self.i1range = (0, np.inf)
        if np.isscalar(ld):
            def constval(x, b, i):
                return x
            self.ld = partial(constval, ld)
            self.psim = partial(constval, psim)
            self.lq = partial(constval, lq)
            logger.debug("ld %s lq %s psim %s", ld, lq, psim)
            return

        if len(ld) == 1:
            try:
                self.io = iqd(min(beta)*np.pi/360, max(i1)/2)
            except:
                self.io = (1, -1)
            def constval(x, b, i):
                return x
            self.ld = partial(constval, ld[0])
            self.psim = partial(constval, psim[0])
            self.lq = partial(constval, lq[0])
            logger.debug("ld %s lq %s psim %s", ld, lq, psim)
            return

        beta = np.asarray(beta)/180.0*np.pi
        if np.any(beta[beta > np.pi]):
            beta[beta > np.pi] = beta - 2*np.pi

        self.io = iqd((np.min(beta)+max(beta))/2, np.max(i1)/2)
        kx = ky = 3
        if len(i1) < 4:
            ky = len(i1)-1
        if len(beta) < 4:
            kx = len(beta)-1
        try:
            pfe = kwargs['losses']
            if 'styoke_exc' in pfe: 
                self.bertotti = True
                self.losskeys += ['styoke_exc', 
                                  'stteeth_exc', 
                                  'rotor_exc']
            self._set_losspar(pfe)
            self._losses = {k: ip.RectBivariateSpline(
                beta, i1, np.array(pfe[k]),
                kx=kx, ky=ky).ev for k in tuple(self.losskeys)}
        except KeyError as e:
            logger.warning("loss map missing: %s", e)
            pass
        if 'psid' in kwargs:
            self.betarange = min(beta), max(beta)
            self.i1range = (0, np.max(i1))
            self.psid = ip.RectBivariateSpline(
                beta, i1, np.sqrt(2)*np.asarray(kwargs['psid']),
                kx=kx, ky=ky).ev
            self.psiq = ip.RectBivariateSpline(
                beta, i1, np.sqrt(2)*np.asarray(kwargs['psiq']),
                kx=kx, ky=ky).ev
            return

        if len(i1) < 4 or len(beta) < 4:
            if len(i1) == len(beta):
                self.ld = ip.interp2d(beta, i1, ld.T)
                self.psim = ip.interp2d(beta, i1, psim.T)
                self.lq = ip.interp2d(beta, i1, lq.T)
                logger.debug("interp2d beta %s i1 %s", beta, i1)
                return
            elif len(i1) == 1:
                def interp(x, b, i):
                    return ip.InterpolatedUnivariateSpline(beta, x, k=1)(b)
                self.ld = partial(interp, ld)
                self.psim = partial(interp, psim)
                self.lq = partial(interp, lq)
                logger.debug("interpolatedunivariatespline beta %s", beta)
                return
            if len(beta) == 1:
                def interp(x, b, i):
                    return ip.InterpolatedUnivariateSpline(i1, x, k=1)(i)
                self.ld = partial(interp, ld)
                self.psim = partial(interp, psim)
                self.lq = partial(interp, lq)
                logger.debug("interpolatedunivariatespline i1 %s", i1)
                return

            raise ValueError("unsupported array size {0}x{1}".format(
                len(beta), len(i1)))

        self.betarange = min(beta), max(beta)
        self.i1range = (0, np.max(i1))
        def interp(x, b, i):
            return ip.RectBivariateSpline(beta, i1, np.asarray(x)).ev(b, i)
        self.ld = partial(interp, ld)
        self.psim = partial(interp, psim)
        self.lq = partial(interp, lq)
        logger.debug("rectbivariatespline beta %s i1 %s", beta, i1)

    def psi(self, iq, id, tol=1e-4):
        """return psid, psiq of currents iq, id"""
        beta, i1 = betai1(np.asarray(iq), np.asarray(id))
        if np.isclose(beta, np.pi, atol=1e-4):
            beta = -np.pi
        #logger.debug('beta %f (%f, %f) i1 %f %f',
        #             beta, self.betarange[0], self.betarange[1],
        #             i1, self.i1range[1])
        if self.check_extrapolation:
            if (self.betarange[0]-tol > beta or
                self.betarange[1]+tol < beta or
                i1 > 1.01*self.i1range[1]):
                return (np.nan, np.nan)
        if self.psid:
            return (self.psid(beta, i1), self.psiq(beta, i1))

        psid = self.ld(beta, i1)*id + np.sqrt(2)*self.psim(beta, i1)
        psiq = self.lq(beta, i1)*iq
        return (psid, psiq)

    def iqdmin(self, i1):
        """max iq, min id for given current"""
        if self.betarange[0] <= -np.pi/2 <= self.betarange[1]:
            return iqd(-np.pi/2, i1)
        if self.betarange[1] == 0:
            return iqd(self.betarange[0], i1)
        return iqd(self.betarange[1], i1)

    def iqdmax(self, i1):
        """max iq, min id for given current"""
        if self.betarange[1] == 0:
            return iqd(self.betarange[1], i1)
        return iqd(self.betarange[0], i1)

    def betai1_plfe1(self, beta, i1, f1):
        stator_losskeys = ['styoke_eddy', 'styoke_hyst',
                            'stteeth_eddy', 'stteeth_hyst']
        if self.bertotti: 
            stator_losskeys += ['styoke_exc', 'stteeth_exc']
        return np.sum([
            self._losses[k](beta, i1)*(f1/self.fo)**self.plexp[k] for
            k in tuple(stator_losskeys)], axis=0)

    def iqd_plfe1(self, iq, id, f1):
        return self.betai1_plfe1(*betai1(iq, id), f1)

    def betai1_plfe2(self, beta, i1, f1):
        rotor_losskeys = ['rotor_eddy', 'rotor_hyst']
        if self.bertotti: 
            rotor_losskeys += ['rotor_exc']
        return np.sum([
            self._losses[k](beta, i1)*(f1/self.fo)**self.plexp[k] for
            k in tuple(rotor_losskeys)], axis=0)

    def iqd_plfe2(self, iq, id, f1):
        return self.betai1_plfe2(*betai1(iq, id), f1)

    def betai1_plmag(self, beta, i1, f1):
        return self._losses['magnet'](beta, i1)*(f1/self.fo)**2

    def iqd_plmag(self, iq, id, f1):
        return self.betai1_plmag(*betai1(iq, id), f1)


class PmRelMachinePsidq(PmRelMachine):
    """Standard set of PM machine parameters:
    p number of pole pairs
    m number of phases

    psid d-flux (Vs Peak)
    psiq q-flux (Vs Peak)
    r1 stator resistance (Ohm)
    r1 stator leakage inductance (H)
    id q current (A, Peak)
    iq q current (A, Peak)
    """

    def __init__(self, m, p, psid, psiq, r1, id, iq, ls=0, **kwargs):
        super(self.__class__, self).__init__(m, p, r1, ls, **kwargs)

        if isinstance(psid, (float, int)):
            self._psid = lambda id, iq: np.array([[psid]])
            self._psiq = lambda id, iq: np.array([[psiq]])
            return

        psid = np.asarray(psid)
        psiq = np.asarray(psiq)
        id = np.asarray(id)
        iq = np.asarray(iq)
        self.idrange = (min(id), max(id))
        self.iqrange = (min(iq), max(iq))
        self.betarange = (-np.pi if min(iq) < 0 else -np.pi/2,
                          0 if max(iq) > 0 else -np.pi/2)
        self.i1range = (0, np.sqrt(2)*np.min(id))
        self.io = np.max(iq)/2, np.min(id)/2

        if np.any(psid.shape < (4, 4)):
            if psid.shape[0] > 1 and psid.shape[1] > 1:
                self._psid = ip.interp2d(iq, id, psid.T)
                self._psiq = ip.interp2d(iq, id, psiq.T)
                return
            if len(id) == 1 or psid.shape[1] == 1:
                def interp(x, q, d):
                    return ip.InterpolatedUnivariateSpline(iq, x)(q)
                self._psid = partial(interp, psid)
                self._psiq = partial(interp, psiq)
                return
            if len(iq) == 1 or psid.shape[0] == 1:
                def interp(x, q, d):
                    return ip.InterpolatedUnivariateSpline(id, x)(d)
                self._psid = partial(interp, psid)
                self._psiq = partial(interp, psiq)
                return
            raise ValueError("unsupported array size {}".format(
                psid.shape))

        self._psid = ip.RectBivariateSpline(iq, id, psid).ev
        self._psiq = ip.RectBivariateSpline(iq, id, psiq).ev
        try:
            pfe = kwargs['losses']
            self._set_losspar(pfe)
            self._losses = {k: ip.RectBivariateSpline(
                iq, id, np.array(pfe[k])).ev for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy',
                'magnet')}
        except KeyError as e:
            logger.warning("loss map missing: %s", e)
            pass

    def psi(self, iq, id):
        return (self._psid(iq, id),
                self._psiq(iq, id))

    def iqdmin(self, i1):
        """max iq, min id for given current"""
        if self.idrange[0] < 0 and self.idrange[1] <= 0:
            idmin = -np.sqrt(2)*i1
        else:
            idmin = 0
        if self.idrange[0] <= idmin/np.sqrt(2):
            iqmin = -np.sqrt(2)*i1
            if self.iqrange[0] <= iqmin:
                return (iqmin, idmin)
            return self.iqrange[0], idmin

        beta = np.arccos(self.iqrange[0]/i1/np.sqrt(2))
        iqmin = np.sqrt(2)*i1*np.sin(beta)
        if self.iqrange[0] <= iqmin:
            return (iqmin, idmin)

        return self.iqrange[0], self.idrange[0]

    def iqdmax(self, i1):
        """max iq, max id for given current"""
        iqmax = np.sqrt(2)*i1
        if iqmax <= np.max(self.iqrange):
            if np.min(self.idrange) < 0 and np.max(self.idrange) <= 0:
                idmax = 0
            else:
                idmax = np.sqrt(2)*i1
            if idmax <= np.max(self.idrange):
                return (iqmax, idmax)
            return (iqmax, np.max(self.idrange))

        beta = np.arccos(self.iqrange[1]/iqmax)
        iqmax = np.sqrt(2)*i1*np.cos(beta)
        idmax = np.sqrt(2)*i1*np.sin(beta)
        if idmax <= np.max(self.idrange):
            return (iqmax, idmax)

        return iqmax, np.max(self.idrange)

    def iqd_plfe1(self, iq, id, f1):
        return np.sum([
            self._losses[k](iq, id)*(f1/self.fo)**self.plexp[k] for
            k in ('styoke_eddy', 'styoke_hyst',
                  'stteeth_eddy', 'stteeth_hyst')], axis=0)

    def betai1_plfe1(self, beta, i1, f1):
        return self.iqd_plfe1(*iqd(beta, i1), f1)

    def iqd_plfe2(self, iq, id, f1):
        return np.sum([
            self._losses[k](iq, id)*(f1/self.fo)**self.plexp[k] for
            k in ('rotor_eddy', 'rotor_hyst',)], axis=0)

    def betai1_plfe2(self, beta, i1, f1):
        return self.iqd_plfe2(*iqd(beta, i1), f1)

    def iqd_plmag(self, iq, id, f1):
        return self._losses['magnet'](iq, id)*(f1/self.fo)**2

    def betai1_plmag(self, beta, i1, f1):
        return self.iqd_plmag(*iqd(beta, i1), f1)



'''Convert Maxwell/msh Files to Gmsh/msh

  Usage:
    mxw2msh <infile> <outfile>

'''
import re
from collections import namedtuple

def mxw2msh(mxfile: str) -> list:
    """convert Maxwell msh file to Gmsh msh

    Args:
        mxfile: the maxwell msh file to convert

    Returns:
        the list of Gmsh msh lines
    """
    with open(mxfile) as f:
        f = iter(f.readlines())

    def next_line(f, n=1):
        for i in range(n):
            line = next(f)
        line = line.strip()
        line = re.split(" |,", line)
        return line

    Node = namedtuple("Node", ["x", "y"])
    nodes = []

    Element = namedtuple("Element", ["n1", "n2", "n3", "entity"])
    elements = []

    while True:
        try:
            line = next_line(f)
            if line[0] == "$begin":

                if line[1] == "'points'":
                    line = next_line(f, 4)

                    while line[0] != "$end":
                        nodes.append(Node(line[3], line[5]))
                        line = next_line(f)

                if line[1] == "'elements'":
                    line = next_line(f)

                    while line[0] != "$end":
                        elements.append(Element(line[3], line[5], line[7], line[9]))
                        line = next_line(f)

        except StopIteration:
            break

    msh = ["$MeshFormat\n2.2 0 8\n$EndMeshFormat"]

    msh.append("$Nodes")
    msh.append(str(len(nodes)))
    for i, n in enumerate(nodes):
        msh.append("{} {} {} 0.0".format(i + 1, n.x, n.y))
    msh.append("$EndNodes")

    msh.append("$Elements")
    msh.append(str(len(elements)))
    for i, e in enumerate(elements):
        msh.append("{} 2 2 0 {} {} {} {}".format(i + 1, e.entity, e.n1, e.n2, e.n3))
    msh.append("$EndElements")
    return msh

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("infile")
    parser.add_argument("outfile")

    args = parser.parse_args()

    infile = args.infile
    outfile = args.outfile

    msh = mxw2msh(infile)
    with open(outfile, "w") as outfile:
        outfile.write("\n".join(msh))



# -*- coding: utf-8 -*-
"""read BCH/BATCH files

"""
import sys
import numpy as np
import re
import logging
import logging.config

logger = logging.getLogger('femagtools.bch')

alpha20 = 3.93e-3  # temperature coeff of copper

_indxOpPattern = re.compile(r'[a-zA-Z0-9_]+(\[[-0-9]+\])')
_statloss = re.compile(r'Fe-Losses\s*Stator')
_rotloss = re.compile(r'Fe-Losses\s*Rotor')


def splitindex(name):
    "returns listname and index if name contains index"
    m = _indxOpPattern.search(name)
    if m:
        return (name.split('[')[0],
                int(''.join(list(m.group(1))[1:-1])))
    return (name, None)


def floatnan(s):
    """converts string to float
    returns NaN on conversion error"""
    try:
        return float(s)
    except ValueError:
        return float('NaN')


def r1_20(r1, theta):
    return r1/(1+alpha20*(theta-20))


def get_si_factor(contentline):
    "extract the first pattern and return conversion factor for SI unit"
    pattern = re.compile(r"\[([A-Za-z/0-9]+)\]")
    search = pattern.search(contentline)
    if search:
        if search.group(1) in ('kW', 'kNm'):
            return 1e3
    return 1.0


def sttylosses(losses):
    """map losses of stator subregions to stteeth and styoke"""
    # extract stator subregion names
    sregs = set([k for k in losses.keys()
                 if k.split('_')[-1] not in {'hyst', 'eddy'} and
                 k not in {'rotor', 'magnet', 'speed'}])

    def hysteddy(yoke, teeth, losses):
        d = {'styoke': losses[yoke]}
        if teeth:
            d['stteeth'] = losses[teeth]
        try:
            d['styoke_hyst'] = losses[yoke+'_hyst']
            d['stteeth_hyst'] = losses[teeth+'_hyst']
            d['styoke_eddy'] = losses[yoke+'_eddy']
            d['stteeth_eddy'] = losses[teeth+'_eddy']
        except KeyError:
            pass
        return d

    if sregs == {'StYoke', 'StTeeth'}:
        return hysteddy('StYoke', 'StTeeth', losses)
    if sregs == {'StJo', 'StZa'}:
        return hysteddy('StJo', 'StZa', losses)
    if sregs == {'Iron', 'StJo'}:
        return hysteddy('StJo', 'Iron', losses)
    if sregs == {'stat'}:
        return hysteddy('stat', '', losses)
    if sregs == {'Iron'}:
        return hysteddy('Iron', '', losses)
    if sregs == {'StJo'}:
        return hysteddy('StJo', '', losses)
    if sregs == {'STTH', 'STTP', 'STYK'}:
        l = hysteddy('STYK', 'STTH', losses)
        l['stteeth'] += losses['STTP']
        try:
            l['stteeth_eddy'] += losses['STTP_eddy']
            l['stteeth_hyst'] += losses['STTP_hyst']
        except KeyError:
            pass
        return l
    return {}


def _readSections(f):
    """return list of bch sections

    sections are surrounded by lines starting with '[***'

    Args:
      param f (file) BCH file to be read

    Returns:
      list of sections
    """

    section = []
    for line in f:
        if line.startswith('[****'):
            if section:
                # skip empty lines
                i = 0
                try:
                    while not section[i]:
                        i = i+1
                except IndexError:
                    i = i-1
                yield section[i:]
                section = []
        else:
            section.append(line.strip())
    yield section


class Reader:
    """Reads a BCH/BATCH-File"""
    _numPattern = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)\s*')
    _numPatternNaN = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?|nan)\s*')

    def __init__(self):
        self._fft = None
        self.type = None
        self.filename = ''
        self.external_rotor = False
        self.project = ''
        self.date = ''
        self.version = ''
        # known types:
        # MULTIPLE CALCULATION OF FORCES AND FLUX
        # Fast cogging calculation OF FORCES AND FLUX
        # Fast LD-LQ-Identification
        # Fast Psid-Psiq-Identification
        # Fast PM-Synchronous-Motor Simulation
        # Characteristics of Permanent-Magnet-Motors
        # T(n) simulation from Ld-Lq-Psim values
        self.wdg = None
        self.wdgfactors = []
        self.torque = []
        self.torque_fft = []
        self.psidq = {}
        self.psidq_ldq = None
        self.machine = {}
        self.lossPar = {}
        self.flux = {}
        self.magnet = {}
        self.airgapInduction = {}
        self.airgap = {}
        self.flux_fft = {}
        self.linearForce = []
        self.linearForce_fft = []
        self.powerSituation = {}
        self.scData = {}
        self.dqPar = {}
        self.ldq = {}
        self.losses = []
        self.magnet_loss_th = []
        self.demag = []
        self.weights = []
        self.weight = {}
        self.characteristics = []
        self.areas = []
        self.current_angles = []
        self.torque_opt = []
        self.dispatch = {
            'General Machine Data': Reader.__read_general_machine_data,
            'Weigths': Reader.__read_weights,
            'Number of Nodes': Reader.__read_nodes_and_mesh,
            'Windings Data': Reader.__read_dummy,
            'Winding-Factors': Reader.__read_winding_factors,
            'Machine Data': Reader.__read_machine_data,
            'CAD-Parameter Data': Reader.__read_dummy,
            'Torque-Force': Reader.__read_torque_force,
            'Airgap Induction Br': Reader.__read_airgapInduction,
            'Fourier Analysis': Reader.__read_fft,
            'Machine excitation': Reader.__read_machine_excitation,
            'DQ-Parameter for open Winding Modell': Reader.__read_dq_parameter,
            'Magnet Data': Reader.__read_magnet_data,
            'Date': Reader.__read_date,
            'Inertia': Reader.__read_inertia,
            'Losses': Reader.__read_dummy,
            'Fe-Hysteresis- and Eddy current Losses[W] > 0.1 % Max':
            Reader.__read_hysteresis_eddy_current_losses,
            'Losses [W]': Reader.__read_losses,
            'Losses for speed [1/min]': Reader.__read_losses_tab,
            'Losses from PSID-Psiq-Identification for speed [1/min]':
            Reader.__read_losses_tab,
            'Magnet loss data': Reader.__read_magnet_loss_data,
            'Project File name': Reader.__read_project_filename,
            'File name': Reader.__read_filename,
            'Windings input data': Reader.__read_windings,
            'Control parameters for Loss calculation': Reader.__read_lossPar,
            'PSID-Psiq-Identification': Reader.__read_psidq,
            'Ld-Lq-Identifikation aus PSID-Psiq-Identification':
            Reader.__read_psidq_ldq,
            'Ld-Lq-Identification': Reader.__read_ldq,
            'Ld-Lq-Identification RMS-values': Reader.__read_ldq,
            'Machine Data Rotor': Reader.__read_dummy,
            'Current Angles defined from no-load test':
            Reader.__read_current_angles,
            'FEMAG Version': Reader.__read_version,
            'FEMAG Classic Version': Reader.__read_version,
            'Simulation Data': Reader.__read_simulation_data,
            'Area [mm**2]': Reader.__read_areas,
            'Basic Machine parameters': Reader.__read_dummy,
            'Winding': Reader.__read_dummy,
            'Calculation time [sec]': Reader.__read_calctime,
            'Results for Angle I-Up [Degree]': Reader.__read_dummy,
            'Demagnetisation': Reader.__read_demagnetization,
            'Transient short circuit': Reader.__read_short_circuit,
            'Peak winding currents': Reader.__read_peak_winding_currents,
            'Flux observed': Reader.__read_flux,
            'Linear Force': Reader.__read_linear_force,
            'Demagnetization Data': Reader.__read_demagnetization,
            'Power Situation (VZS)': Reader.__read_power_situation,
            '** Characteristics of Permanent-Magnet-Motors **':
            Reader.__read_characteristics}

    def getStep(self):
        """@returns displacement step of flux values"""
        if len(self.flux) > 0:
            return self.flux[0]['displ'][1]-self.flux[0]['displ'][0]
        return None

    def read(self, content):
        """read bch file

        Args:
          content (str or list of str) the text lines of the BCH file
        """
        if isinstance(content, str):
            lines = content.split('\n')
        else:
            lines = content
        for s in _readSections(lines):
            if not s:
                continue
            title = s[0].split(':')[0].strip()
            if title == 'Function':
                title = s[0].split(':')[1].strip()
            logger.debug("'%s': %d", title, len(s[1:]))
            # Check if we are finished with the fourier analysis part(s)
            if title != 'Fourier Analysis':
                self._fft = None
            else:
                title2 = s[0].split(':')[1].strip()
                for k in ['Airgap Induction Br']:
                    if k == title2[:len(k)]:
                        title = title2[:len(k)]
            if title in self.dispatch:
                self.dispatch[title](self, s)

        if len(self.weights) > 0:
            w = list(zip(*self.weights))
            self.weight['iron'] = sum(w[0])
            self.weight['conductor'] = sum(w[1])
            self.weight['magnet'] = sum(w[2])
            self.weight['total'] = sum([sum(l) for l in w])
        return self

    def __findNums(self, l):
        rec = self._numPattern.findall(l)
        if 3 * '*' in l:  # min 3 '*'
            li = re.sub(r'[\*]+', str(np.NaN), l)
            rec = self._numPatternNaN.findall(li)
            logger.debug("Numbers with **** In: %s", l)
            logger.debug("Out: %s", li)
            logger.debug("%d. NUMS: %s", len(rec), rec)
        return rec

    def __read_version(self, content):
        rec = content[0].split(':')
        self.version = rec[-1].replace(' Version ', '').strip()

    def __read_project_filename(self, content):
        self.project = content[1].strip()

    def __read_filename(self, content):
        self.filename = content[0].split(':')[1].strip()

    def __read_date(self, content):
        d = content[0].split(':')[1].strip().split()
        dd, MM, yy = d[0].split('.')
        hh, mm = ''.join(d[1:-1]).split('.')
        self.date = '{}-{}-{}T{:02}:{:02}'.format(yy, MM, dd, int(hh), int(mm))

    def __read_nodes_and_mesh(self, content):
        self.nodes, self.elements, self.quality = \
            [floatnan(r[0]) for r in [self.__findNums(l)
                                      for l in content[:3]]]
        for l in content[3:]:
            m = re.match(r'\*+([^\*]+)\*+', l)
            if m:
                self.type = m.group(1).strip()
                return

    def __read_magnet_data(self, content):
        """read magnet data section"""
        for l in content:
            for v in [["Remanence Br [T]", 'Br'],
                      ["Co. Force Hc", 'Hc'],
                      ["Rel. Permeability", 'muer'],
                      ["Magn.Temperatur[grad]", 'Tmag'],
                      ["Temp.coeficient Br", 'alpha'],
                      ["Limit demagnetization   [%]", 'demag_pc'],
                      ["Limit demagnetization Hx [kA/m]", 'demag_hx'],
                      ["Total magnet area", 'area']]:

                if l.startswith(v[0]):
                    rec = self.__findNums(l)
                    if len(rec) > 0:
                        self.magnet[v[1]] = floatnan(rec[-1])
                    break

    def __read_windings(self, content):
        "read winding section"
        index = False
        wdgs = []
        for line in content:
            rec = line.split()
            if len(rec) > 0 and not index:
                index = rec[0].startswith('Index')
            elif len(rec) == 5 and index:
                if rec[0].startswith('Number'):
                    index = False
                else:
                    wdgs.append((abs(int(rec[1])),
                                 1 if int(rec[1]) > 0 else -1,
                                 float(rec[2]), float(rec[3]),
                                 float(rec[4])))
            else:
                index = False
        phases = set(list(zip(*wdgs))[0])
        self.windings = dict([(k, dict([(j, [])
                                        for j in ('dir',
                                                  'N',
                                                  'R',
                                                  'PHI')]))
                              for k in phases])
        for w in wdgs:
            self.windings[w[0]]['dir'].append(w[1])
            self.windings[w[0]]['N'].append(w[2])
            self.windings[w[0]]['R'].append(w[3]/1e3)
            self.windings[w[0]]['PHI'].append(w[4])

    def __read_winding_factors(self, content):
        "read winding factors section"
        self.wdgfactors = []
        for line in content:
            if line.startswith('Winding-Key'):
                self.wdgfactors.append(dict(order=[], wfac=[],
                                            skewf=[], total=[]))
            else:
                rec = self.__findNums(line)
                if len(rec) == 4:
                    self.wdgfactors[-1]['order'].append(int(rec[0]))
                    self.wdgfactors[-1]['wfac'].append(float(rec[1]))
                    self.wdgfactors[-1]['skewf'].append(float(rec[2]))
                    self.wdgfactors[-1]['total'].append(float(rec[3]))

    def __read_dummy(self, content):
        return

    def __read_magnet_loss_data(self, content):
        for line in content:
            if line.startswith('El.Conductivity Magnet'):
                try:
                    self.magnet['sigma_PM'] = float(line.split()[-1])
                except ValueError:
                    pass

    def __read_calctime(self, content):
        try:
            self.calctime = float(content[1])
        except (IndexError, ValueError):
            pass

    def __read_simulation_data(self, content):
        for i, line in enumerate(content):
            if line.startswith('Number of Phases m'):
                self.machine['m'] = int(float((line.split()[-1])))
            elif (line.startswith('Radius air-gap center (torque)') or
                  line.startswith('Air-gap center 1:radius/position')):
                rec = [l.strip() for l in line.split()]
                fc_radius = 1e-3*float(rec[-1])
                self.machine['fc_radius'] = fc_radius
            elif line.startswith('Number of parallel Windings'):
                self.machine['num_par_wdgs'] = int(float(line.split()[-1]))
            elif line.startswith('leak_dist_wind'):
                self.leak_dist_wind = self.__read_leak_dist_wind(content[i+1:])
        return

    def __read_leak_dist_wind(self, content):
        leak_wind = dict()
        vmap = {
            'Number of segments': 'nseg',
            'Number of poles simulated': 'npolsim',
            'Radius air-gap center (torque) [mm]': 'fc_radius',
            'Radius of perimeter [mm]': 'perimrad',
            'Bending radius vertical [mm]': 'vbendrad',
            'End winding height [mm]': 'endheight',
            'Wire diameter [mm]': 'wiredia',
            'Armature length [mm]': 'armatureLength',
            'End winding length [mm]': 'wdgendlen',
            'External inductance L0e/winding [H]': 'L0e',
            'Lde/winding [H]': 'Lde',
            'Lqe/winding [H]': 'Lqe',
            'Internal inductance L0i/winding [H]': 'L0i',
            'Ldi/winding [H]': 'Ldi',
            'Lqi/winding [H]': 'Lqi'}

        for line in content:
            rec = [l.strip() for l in line.split()]
            try:
                k = vmap[' '.join(rec[:-1])]
                v = float(rec[-1])
                if rec[-2] == '[mm]':
                    leak_wind[k] = 1e-3*v
                elif rec[-2] == '[H]':
                    leak_wind[k] = v
                else:
                    leak_wind[k] = int(v)
            except (KeyError, ValueError):
                pass
        return leak_wind

    def __read_current_angles(self, content):
        self.current_angles = []
        r = []
        extracting = False
        for l in content:
            rec = self.__findNums(l)
            if len(rec) == 3:
                self.current_angles.append(floatnan(rec[-1]))
            if l.strip().startswith('Optimization of angle'):
                extracting = True
            elif l.strip().startswith('Optimimal angle I'):
                break
            else:
                if extracting:
                    if l.strip() == '' or \
                        l.strip().startswith('Angle [Degr'):
                        pass
                    else:
                        tmp = l.split('\t')
                        r.append([floatnan(k) for k in tmp])
        if len(r) > 1:
            self.torque_opt.append({'beta': [i[0] for i in r],
                                    'torque': [i[1] for i in r]})
        return

    def __read_machine_excitation(self, content):
        rec = self.__findNums(content[1])
        if len(rec) == 1:
            if content[1].startswith('Magnets:'):
                self.machine['brem'] = floatnan(rec[-1])
            else:
                self.machine['ex_current'] = floatnan(rec[-1])
        return

    def __read_lossPar(self, content):
        self.lossPar = {
            'fo': [],
            'Bo': [],
            'ch': [],
            'cw': [],
            'hf': [],
            'ef': [],
            'ic': [],
            'gamfe': [],
            'thetaw': [],
            'fillfe': [],
            'culength': [],
            'mat': []}
        for l in content:
            for v in [["Base Frequency", 'fo'],
                      ["Base Induction", 'Bo'],
                      ["Hysteresis-Coefficient", 'ch'],
                      ["Eddycurrent-Coefficient", 'cw'],
                      ["Hysteresis-Frequency-Coefficient", 'hf'],
                      ["Eddycurrent-Frequency-Coefficient", 'ef'],
                      ["Induction-Coefficient", 'ic'],
                      ["Specific Weight Iron", 'gamfe'],
                      ["Conductor Temperature", 'thetaw'],
                      ["Fillfactor Iron", 'fillfe'],
                      ["Relative conductor length", 'culength'],
                      ["Material factor", 'mat']]:

                if l.find(v[0]) > -1:
                    rec = self.__findNums(l)
                    if len(rec) > 0:
                        if v[1] == 'culength':
                            self.lossPar[v[1]].append(floatnan(rec[-1])/100)
                        else:
                            self.lossPar[v[1]].append(floatnan(rec[-1]))
                    break

        return l

    def __read_demagnetization(self, content):
        keys = [('displ', 'current_1', 'current_2', 'current_3',
                 'H_max', 'H_av', 'area'),
                ('displ', 'current', 'beta',
                 'H_max', 'H_av', 'area')]

        demag = dict()
        for l in content:
            rec = self.__findNums(l)
            if len(rec) == 7 or len(rec) == 6:
                i = 0 if len(rec) == 7 else 1
                self.demag.append(
                    {k: floatnan(r) for r, k in zip(rec, keys[i])})

                if demag and 'segment' in demag:
                    self.demag[-1].update(demag)
            else:
                for v in [["Limit Hc value", "lim_hc"],
                          ["Max. Magnetization", "br_max"],
                          ["Demagnetisation", "segment"],
                          ["Min. Magnetization", "br_min"],
                          ["Area demagnetized", "area"]]:
                    if l.find(v[0]) > -1:
                        if len(rec) > 0:
                            if v[1] == 'segment':
                                demag[v[1]] = int(rec[-1])
                            else:
                                demag[v[1]] = floatnan(rec[-1])
                        break
        if demag:
            if not 'segment' in demag:
                self.demag.append(demag)

    def __read_short_circuit(self, content):
        "read short circuit section"
        if content[2].startswith('Time'):
            m = []
            for l in content:
                rec = l.split()
                if len(rec) == 5 and not rec[0].startswith('Time'):
                    m.append([floatnan(x) for x in rec])

            m = np.array(m).T
            self.scData['time'] = m[0].tolist()
            self.scData['ia'] = m[1].tolist()
            self.scData['ib'] = m[2].tolist()
            self.scData['ic'] = m[3].tolist()
            self.scData['torque'] = m[4].tolist()
            return

        l = content[-1]
        rec = l.split()
        self.scData['speed'] = floatnan(rec[0])
        self.scData['ikd'] = floatnan(rec[1])
        self.scData['tkd'] = floatnan(rec[2])
        self.scData['iks'] = floatnan(rec[3])
        self.scData['tks'] = floatnan(rec[4])

    def __read_peak_winding_currents(self, content):
        self.scData['peakWindingCurrents'] = [float(x)
                                              for x in re.findall(r'[-0-9.]+',
                                                                  ''.join(content))]

    def __read_general_machine_data(self, content):
        mcfiles = []
        for l in content:
            try:
                if l.find('Armature Length') > -1:
                    self.armatureLength = floatnan(l.split()[-1])
                elif l.find('No of parallel wdgs-section') > -1:
                    self.machine['parallelWdgs'] = int(l.split()[-1])
                elif l.find('Skew angle') > -1:
                    self.machine['skewAngle'] = floatnan(l.split()[-1])
                elif l.find('number of skew sections') > -1:
                    self.machine['skewSections'] = int(l.split()[-1])
                elif l.find('Magn. Fluss Psim RMS') > 0:
                    self.machine['psim'] = floatnan(l.split()[-1])
                elif l.find('Number of Pole pairs') > -1:
                    self.machine['p'] = int(l.split()[-1])
                elif l.find('Number of Poles simulated') > -1:
                    self.machine['p_sim'] = int(l.split()[-1])
                elif l.find('Total Number of Slots') > -1:
                    self.machine['Q'] = int(l.split()[-1])
                elif l.find('Number of Slot-Sides sim.') > -1:
                    self.machine['qs_sim'] = int(l.split()[-1])
                elif l.find('POC-File used in calculation') > -1:
                    self.machine['pocfile'] = l.split(
                        ':')[-1].strip().replace('\\', '\\\\')
                elif l.find('MC-File used in calculation') > -1:
                    mcfiles.append(l.split()[-1].strip().replace('\\', '\\\\'))
                elif l.find('Rotation Fraction') > -1:
                    self.machine['period_frac'] = int(l.split()[-1])
            except Exception as e:
                logger.warning('%s: %s', e, l)
        if mcfiles:
            if len(mcfiles) > 1:
                self.machine['mcfile'] = mcfiles
            else:
                self.machine['mcfile'] = mcfiles[0]

    def __read_characteristics(self, content):
        characteristics = {}
        for i, l in enumerate(content):
            if l.startswith('[[***'):
                break
            for v in [['Voltage (operat. limit)', 'u1nom'],
                      ['Current (operat. limit)', 'i1nom'],
                      ['Angle I VS Up (speed = 0)', 'beta0'],
                      ['Resistance stator winding', 'r1'],
                      ['Inductance(I,Angle I-Up)', 'Ldnom'],
                      ['Inductance(I,Angle I-Up)', 'Lqnom'],
                      ['Power (operating limit)',  'Pnom'],
                      ['Stator ewdg inductance',   'Le'],
                      ['Stator external inductance Lex', 'Lex'],
                      ['Magn. flux (RMS)', 'psimnom'],
                      ['Effect. armature length', 'lfe'],
                      ['Power cut-off speed NC', 'nc'],
                      ['Number of Pole pairs', 'p'],
                      ['Max. current (RMS)', 'i1max'],
                      ['Rel. number wdg turns(wdg.1)', 'relw'],
                      ['Number of Phases', 'm'],
                      ['Min. speed', 'nmin'],
                      ['Max. speed', 'nmax']]:

                if l.find(v[0]) > -1:
                    rec = self.__findNums(l)
                    if len(rec) > 0:
                        characteristics[v[1]] = floatnan(rec[-1])
                        break

        characteristics['ldq'] = {}
        m = []
        if content[i+1].startswith('Wdg'):
            for k, l in enumerate(content[i+3:]):
                if l.startswith('[[***'):
                    break
                rec = l.split('\t')
                if len(rec) == 6:
                    m.append([floatnan(x) for x in rec])
        else:
            k = -3
        if m:
            m = np.array(m).T
            ncols = len(set(m[1]))
            i1 = np.reshape(m[0], (-1, ncols)).T[0]
            nrows = len(i1)

            logger.info('characteristics ld-lq %d x %d', nrows, ncols)
            characteristics['ldq'] = {
                'beta': m[1][:ncols][::-1].tolist(),
                'i1': i1.tolist(),
                'ld': (characteristics['lfe']*np.reshape(
                    m[2], (nrows, ncols)).T[::-1]).tolist(),
                'lq': (characteristics['lfe']*np.reshape(
                    m[3], (nrows, ncols)).T[::-1]).tolist(),
                'psim': (characteristics['lfe']*np.reshape(
                    m[4], (nrows, ncols)).T[::-1]).tolist(),
                'torque': (characteristics['lfe']*np.reshape(
                    m[5], (nrows, ncols)).T[::-1]).tolist()}

        m = []
        columns = [['n', 'id', 'iq', 'torque', 'p2'],
                   ['beta', 'cos_phi', 'u1', 'um'],
                   ['lang', 'ud', 'uq', 'i1'],
                   ['lang', 'ld', 'lq', 'psim']]
        nsec = 0
        characteristics['speed_torque'] = {}
        for l in content[k+i+6:]:
            if l.startswith('[[***'):
                break
            if not l:
                continue
            if l.startswith('Speed') and m:
                if nsec == 0:
                    m = np.array(m).T
                else:
                    m = np.array(m).T[1:]
                for j, k in enumerate(columns[nsec]):
                    characteristics['speed_torque'][k] = m[j].tolist()
                m = []
                nsec += 1
            else:
                rec = self.__findNums(l)
                if len(rec) > 3:
                    m.append([floatnan(x) for x in rec])

        self.characteristics.append(characteristics)

    def __read_flux(self, content):
        "read and append flux section"

        f = {'displ': [], 'flux_k': [], 'voltage_dpsi': [],
             'voltage_four': [], 'current_k': [], 'voltage_ir': []}

        for l in content:
            rec = l.split()
            if l.startswith('Flux-Area'):
                areas = self.__findNums(l)
                if not areas:
                    continue

                self.wdg = areas[0] if len(areas) == 1 else '{}-{}'.format(
                    areas[0], areas[1])
                if self.wdg not in self.flux:
                    self.flux[self.wdg] = []
            elif len(rec) == 7:
                f['displ'].append(floatnan(rec[1].strip()))
                f['flux_k'].append(floatnan(rec[2].strip()))
                f['voltage_dpsi'].append(floatnan(rec[3].strip()))
                f['voltage_four'].append(floatnan(rec[4].strip()))
                f['current_k'].append(floatnan(rec[5].strip()))
                f['voltage_ir'].append(floatnan(rec[6].strip()))
            elif rec and rec[0].startswith('['):
                f['displunit'] = re.search(r"\[([^\]]*)\]", l).group(1).strip()

        self.flux[self.wdg].append(f)
        self._fft = Reader.__read_flux_fft

    def __read_linear_force(self, content):
        "read and append linear force section"
        cosys = 'xy'
        f = {'displ': [], 'magnet_1': [], 'force_x': [],
             'force_y': [], 'f_idpsi': []}

        for l in content:
            rec = self.__findNums(l)
            if len(rec) > 4:
                f['displ'].append(floatnan(rec[1].strip()))
                f['magnet_1'].append(floatnan(rec[2].strip()))
                f['force_x'].append(floatnan(rec[3].strip()))
                f['force_y'].append(floatnan(rec[4].strip()))
                # TODO f['f_idpsi'].append(floatnan(rec[5].strip()))
            elif l.split()[-1] == 'Force_Z':
                cosys = 'rz'

        if cosys == 'rz':
            f['force_r'] = f.pop('force_x')
            f['force_z'] = f.pop('force_y')

        if len(f['displ']) > 0:
            if cosys == 'xy':
                ripple = [max(f['force_x']) - min(f['force_x']),
                          max(f['force_y']) - min(f['force_y'])]
                f['ripple_x'] = ripple[0]
                f['ripple_y'] = ripple[1]
            else:
                ripple = [max(f['force_r']) - min(f['force_r']),
                          max(f['force_z']) - min(f['force_z'])]
                f['ripple_r'] = ripple[0]
                f['ripple_z'] = ripple[1]

            self.linearForce.append(f)

            self._fft = Reader.__read_linearForce_fft

    def __read_linearForce_fft(self, content):
        "read and append linear force fft section"
        if not self._fft:
            return
        linearForce_fft = dict(order=[], force=[], force_perc=[],
                               a=[], b=[])
        for l in content:
            rec = self.__findNums(l)
            if len(rec) > 2:
                linearForce_fft['order'].append(int(rec[0].strip()))
                linearForce_fft['force'].append(floatnan(rec[1].strip()))
                linearForce_fft['force_perc'].append(floatnan(rec[2].strip()))
                if len(rec) > 4:
                    linearForce_fft['a'].append(floatnan(rec[3].strip()))
                    linearForce_fft['b'].append(floatnan(rec[4].strip()))
                else:
                    linearForce_fft['a'].append(0.0)
                    linearForce_fft['b'].append(0.0)

        if linearForce_fft['order']:
            self.linearForce_fft.append(linearForce_fft)

    def __read_fft(self, content):
        if self._fft:
            self._fft(self, content)

    def __read_flux_fft(self, content):
        "read and append flux fft section"

        flux_fft = dict(order=[], flux=[], flux_perc=[],
                        voltage=[], voltage_perc=[], a=[], b=[])
        for l in content:
            rec = self.__findNums(l)
            if len(rec) > 4:
                flux_fft['order'].append(int(rec[0].strip()))
                flux_fft['flux'].append(floatnan(rec[1].strip()))
                flux_fft['flux_perc'].append(floatnan(rec[2].strip()))
                flux_fft['voltage'].append(floatnan(rec[3].strip()))
                flux_fft['voltage_perc'].append(floatnan(rec[4].strip()))
                if len(rec) > 3:
                    flux_fft['a'].append(floatnan(rec[3].strip())*1e-2)
                    flux_fft['b'].append(floatnan(rec[4].strip())*1e-2)
                else:
                    flux_fft['a'].append(0.0)
                    flux_fft['b'].append(0.0)
        if self.wdg not in self.flux_fft:
            self.flux_fft[self.wdg] = []
        self.flux_fft[self.wdg].append(flux_fft)

    def __read_airgapInduction(self, content):
        "read and append airgapInduction section"
        import scipy.integrate as si
        import math

        logger.debug('read airgapInduction')
        i1beta = False  # format is either i1/beta or id/iq
        if 'i1' in self.ldq and 'beta' in self.ldq:
            i1 = self.ldq['i1']
            beta = self.ldq['beta']
        elif 'id' in self.psidq and 'iq' in self.psidq:
            id = self.psidq['id']
            iq = self.psidq['iq']
        else:
            i1 = []
            beta = []
            id = []
            iq = []

        an = [[], [], [], []]
        bn = [[], [], [], []]
        Bm = []
        Ba = []

        for line in content[5:]:
            if line.startswith('[****'):
                break
            if line.startswith("Current"):
                i1beta = True
                continue
            if line.startswith("C_STEP"):
                return   # ignore this section
            try:
                rec = self.__findNums(line)
                if len(rec) == 10:
                    f = [float(s) for s in rec]
                    an[0].append(f[2])
                    bn[0].append(f[3])
                    an[1].append(f[4])
                    bn[1].append(f[5])
                    an[2].append(f[6])
                    bn[2].append(f[7])
                    an[3].append(f[8])
                    bn[3].append(f[9])

                    a = (an[0][-1], an[1][-1], an[2][-1], an[3][-1])
                    b = (bn[0][-1], bn[1][-1], bn[2][-1], bn[3][-1])

                    def B(x):
                        return sum(a[i] * np.cos((2 * i + 1) * x) +
                                   b[i] * np.sin((2 * i + 1) * x)
                                   for i in (0, 1, 2, 3) if not math.isnan(a[i]) and not math.isnan(b[i]))

                    def Bdc(x):
                        return abs(B(x))

                    Ba.append(si.quad(Bdc, 0, 2 * np.pi,
                                      limit=250)[0] / (2 * np.pi))
                    Bm.append(max([B(x) for x in np.linspace(
                        0, 2 * np.pi, 100)]))

            except Exception as e:
                logger.debug("Conversion error: {} :: {}".format(e, line))

        self.airgapInduction = dict()

        if i1beta:
            ncols = len(beta)
            if ncols:
                nrows = len(Ba)//ncols
            else:
                nrows = len(i1)
            self.airgapInduction['beta'] = beta
            self.airgapInduction['i1'] = i1[:nrows]
        else:
            ncols = len(iq)
            self.airgapInduction['iq'] = iq
            self.airgapInduction['id'] = id
            nrows = len(self.airgapInduction['id'])

        try:
            self.airgapInduction['an'] = [np.reshape(an[j][:nrows*ncols],
                                                     (nrows, ncols)).T.tolist()
                                          for j in (0, 1, 2, 3)]
            self.airgapInduction['bn'] = [np.reshape(bn[j][:nrows*ncols],
                                                     (nrows, ncols)).T.tolist()
                                          for j in (0, 1, 2, 3)]
            self.airgapInduction['Bm'] = np.reshape(Bm[:nrows*ncols],
                                                    (nrows, ncols)).T.tolist()
            self.airgapInduction['Ba'] = np.reshape(Ba[:nrows*ncols],
                                                    (nrows, ncols)).T.tolist()
            # check for nan:
            if len(self.airgapInduction['an'][0]) > 1 and \
               len(self.airgapInduction['an'][0][0]) != len(self.airgapInduction['an'][0][1]):
                self.airgapInduction['an'] = [self.airgapInduction['an'][i][1:]
                                              for i in range(3)]
                self.airgapInduction['bn'] = [self.airgapInduction['bn'][i][1:]
                                              for i in range(3)]
                self.airgapInduction['Ba'] = self.airgapInduction['Ba'][1:]
                self.airgapInduction['Bm'] = self.airgapInduction['Bm'][1:]

                if len(self.airgapInduction['an'][0]) > 1 and \
                   len(list(filter(lambda x: np.isnan(x),
                                   list(zip(*self.airgapInduction['an'][0]))[0]))) > 0:
                    self.airgapInduction['an'] = [self.airgapInduction['an'][i][1:]
                                                  for i in range(3)]
                    self.airgapInduction['bn'] = [self.airgapInduction['bn'][i][1:]
                                                  for i in range(3)]
                    self.airgapInduction['Ba'] = zip(*zip(*self.airgapInduction['Ba'])
                                                     [1:])
                    self.airgapInduction['Bm'] = zip(*zip(*self.airgapInduction['Bm'])
                                                     [1:])
        except ValueError:
            print(self.airgapInduction['i1'])

    def __read_torque_force(self, content):
        "read and append force/torque section"

        torque = {
            'angle': [],
            'current_1': [],
            'force_x': [],
            'force_y': [],
            't_idpsi': [],
            'torque': []}
        for l in content:
            rec = self.__findNums(l)
            if len(rec) == 7:
                torque['angle'].append(floatnan(rec[1].strip()))
                torque['current_1'].append(floatnan(rec[2].strip()))
                torque['force_x'].append(floatnan(rec[3].strip()))
                torque['force_y'].append(floatnan(rec[4].strip()))
                torque['t_idpsi'].append(floatnan(rec[5].strip()))
                torque['torque'].append(floatnan(rec[6].strip()))

        if len(torque['angle']) > 0:
            ripple = max(torque['torque']) - min(torque['torque'])
            torque['ripple'] = ripple
            self.torque.append(torque)
        self._fft = Reader.__read_torque_force_fft

    def __read_torque_force_fft(self, content):
        "read and append force/torque fft section"
        columns = content[3].split()
        if len(columns) > 1 and columns[1] == 'Torque':
            torque_fft = dict(order=[],
                              torque=[],
                              torque_perc=[],
                              a=[],
                              b=[])
            for l in content:
                rec = self.__findNums(l)
                if len(rec) > 2:
                    torque_fft['order'].append(int(rec[0].strip()))
                    torque_fft['torque'].append(floatnan(rec[1].strip()))
                    torque_fft['torque_perc'].append(floatnan(rec[2].strip()))
                    if len(rec) > 3:
                        torque_fft['a'].append(floatnan(rec[3].strip())*1e-2)
                        torque_fft['b'].append(floatnan(rec[4].strip())*1e-2)
                    else:
                        torque_fft['a'].append(0.0)
                        torque_fft['b'].append(0.0)
            if torque_fft['order']:
                self.torque_fft.append(torque_fft)

    def __read_power_situation(self, content):
        "read and append power situation section"
        ps = dict()
        beta = ''
        for l in content:
            rec = self.__findNums(l)
            if len(rec) == 7:
                w = rec[0]
                ps[w] = dict(
                    voltage=floatnan(rec[1].strip()),
                    current_1=floatnan(rec[2].strip()),
                    beta=floatnan(rec[3].strip()),
                    cosphi=floatnan(rec[4].strip()),
                    powerp=1e3*floatnan(rec[5].strip()),
                    powerq=1e3*floatnan(rec[6].strip()))
            elif l.startswith('Angle current'):
                beta = rec[0]
        if ps:
            self.powerSituation = dict(windings=ps)
            if beta:
                self.powerSituation['beta'] = float(beta)

    def __removeTrailingZero(self, idList):
        '''if id list is inhomogeneous, remove a trailing '0'
        e.g. : idList[-450, -350, -250, -150, -50, 0]
               idList[-500, -400, -300, -200, -100, 0, 0]
        '''
        if idList[-1] == 0 and len(idList) > 2 and \
           int(idList[-1] - idList[0]) / (len(idList)-1) != \
           int(idList[-2] - idList[0]) / (len(idList)-2):
            idList = idList[:-1]
        return idList

    def __read_psidq(self, content):
        "read psid-psiq section"
        for i, l in enumerate(content):
            if l.find('[A') > -1:
                break
        m = []
        for l in content[i+2:]:
            rec = l.split('\t')
            if len(rec) == 7:
                m.append([floatnan(x) for x in rec])

        m = np.array(m).T
        ncols = np.argmax(np.abs(m[1][1:]-m[1][:-1]))+1
        if ncols == 1 and len(m[1]) > 1 and m[1][0] != m[1][1]:  # simple correction
            ncols = 2
        iq = np.reshape(m[1], (-1, ncols))[0]
        if ncols > 1 and (iq[ncols-1] < iq[ncols-2] or
                          len(m[0]) % ncols != 0):
            ncols = ncols-1

        id = np.reshape(m[0], (-1, ncols)).T[0]
        id = self.__removeTrailingZero(id)
        nrows = len(id)
        if nrows > 1 and id[nrows-1] < id[nrows-2]:
            nrows = nrows-1
        logger.info('psid-psiq %d x %d', nrows, ncols)
        mlen = nrows*ncols
        self.psidq = {k: (self.armatureLength*np.reshape(
            v[:mlen], (nrows, ncols))).T.tolist()
            for k, v in zip(('psid', 'psiq', 'torque_fe', 'torque'),
                            m[3:])}
        self.psidq['iq'] = iq[:ncols].tolist()
        self.psidq['id'] = id[:nrows].tolist()

    def __read_psidq_ldq(self, content):
        "read ldq from psid-psiq section"
        for i, l in enumerate(content):
            if l.find('[A') > -1:
                break
        m = []
        for l in content[i+2:]:
            rec = l.split('\t')
            if len(rec) == 7:
                m.append([floatnan(x) for x in rec])

        m = np.array(m).T
        ncols = np.argmax(np.abs(m[1][1:]-m[1][:-1]))+1
        if ncols == 1 and len(m[1]) > 1 and m[1][0] != m[1][1]:  # simple correction
            ncols = 2
        iq = np.linspace(np.min(m[1]), np.max(m[1]), ncols)
        if ncols > 1 and (iq[ncols-1] < iq[ncols-2] or
                          len(m[0]) % ncols != 0):
            ncols = ncols-1

        id = np.reshape(m[0], (-1, ncols)).T[0]
        id = self.__removeTrailingZero(id)
        nrows = len(id)
        if nrows > 1 and id[nrows-1] < id[nrows-2]:
            nrows = nrows-1
        logger.info('psid-psiq ldq %d x %d', nrows, ncols)
        mlen = nrows*ncols
        self.psidq_ldq = {
            'iq': iq[:ncols].tolist(),
            'id': id[:nrows].tolist(),
            'ld': (self.armatureLength*np.reshape(
                m[2][:mlen], (nrows, ncols)).T).tolist(),
            'lq': (self.armatureLength*np.reshape(
                m[3][:mlen], (nrows, ncols)).T).tolist(),
            'psim': (self.armatureLength*np.reshape(
                m[4][:mlen], (nrows, ncols)).T).tolist(),
            'torque': (self.armatureLength*np.reshape(
                m[6][:mlen], (nrows, ncols)).T).tolist()}

    def __read_ldq(self, content):
        "read ld-lq section"
        keys = ['ld', 'lq', 'psim', 'psid', 'psiq', 'torque']
        for i, l in enumerate(content):
            if l.find('Wdg.-Curr') > -1:
                if l.find('Voltage Up') > -1:
                    keys[2] = 'up'
            if l.find('[A') > -1:
                break
        m = []
        k = i+2
        for l in content[i+2:]:
            rec = l.split('\t')
            if len(rec) > 7:
                m.append([floatnan(x) for x in rec[:8]])
            elif rec and rec[0].startswith('Curr Id'):
                break
            k += 1

        m = np.array(m).T
        ncols = len(set(m[1]))
        i1 = np.reshape(m[0], (-1, ncols)).T[0]
        nrows = len(i1)
        logger.info('ld-lq %d x %d', nrows, ncols)

        self.ldq = {k: (self.armatureLength*np.reshape(
            v, (nrows, ncols)).T[::-1]).tolist() for k, v in zip(
                keys, m[2:])}
        self.ldq['beta'] = m[1][:ncols][::-1].tolist()
        self.ldq['i1'] = i1.tolist()

        # skip d-q table
        i = k+3
        for l in content[k+3:]:
            if l.startswith('Losses for'):
                self.__read_losses_tab(content[i:])
            i += 1

    def __read_losses_tab(self, content):
        "read losses of psidq or ldq"
        m = []
        subregs = [name.split()[-1]
                   for name in content[2].split('\t') if name][2:]
        logger.info("Stator Subregions: %s", subregs)
        speed = float(content[0].split()[-1])/60.
        logger.info('losses for speed %f', speed)
        nl = 4
        for l in content[4:]:
            rec = l.split('\t')
            if len(rec) > 1:
                if rec[0].startswith('P fe'):
                    break
                m.append([floatnan(x) for x in rec])
            nl += 1
        if not m:
            return
        m = np.array(m).T
        ncols = np.argmax(np.abs(m[1][1:]-m[1][:-1]))+1
        if ncols == 1 and len(m[1]) > 1 and m[1][0] != m[1][1]:
            ncols = 2
        id = np.reshape(m[0], (-1, ncols)).T[0]
        id = self.__removeTrailingZero(id)
        nrows = len(id)

        cols = subregs[:-2] + ['rotor', 'magnet']
        mlen = nrows*ncols
        if self.ldq:
            ls = {k: np.reshape(v[:mlen],
                                (nrows, ncols)).T[::-1].tolist()
                  for k, v in zip(cols, m[2:])}
        else:
            ls = {k: np.reshape(v[:mlen],
                                (nrows, ncols)).T.tolist()
                  for k, v in zip(cols, m[2:])}
        m = []
        for l in content[nl+3:]:
            rec = l.split('\t')
            if not rec and m:
                break
            if len(rec) > 5:
                m.append([floatnan(x) for x in rec])

        cols = []
        for s in subregs[:-2] + ['rotor', 'magnet']:
            cols += [s+'_hyst', s+'_eddy']
        if m:
            m = np.array(m).T
            if self.ldq:
                ls.update({k: np.reshape(v[:mlen],
                                         (nrows, ncols)).T[::-1].tolist()
                           for k, v in zip(cols, m[2:])})
            else:
                ls.update({k: np.reshape(v[:mlen],
                                         (nrows, ncols)).T.tolist()
                           for k, v in zip(cols, m[2:])})
        ls['speed'] = speed
        ls.update(sttylosses(ls))
        if self.ldq:
            self.ldq['losses'] = ls
        elif self.psidq:
            self.psidq['losses'] = ls

    def __read_machine_data(self, content):
        "read machine data section"
        for k in ('beta', 'plfe1', 'plfe2', 'plmag', 'plcu'):
            self.machine[k] = []
        for l in content:
            contentline = l.strip()
            if contentline.startswith('Number of phases'):
                rec = l.split()
                self.machine['m'] = int(rec[-1])
                continue
            if contentline.startswith("Mechanical Power"):
                rec = l.split()
                self.machine['p2'] = floatnan(rec[-1])
                unit = rec[2]
                if unit == '[kW]':
                    self.machine['p2'] *= 1e3
                continue

            for v in [["Phase Current", 'i1'],
                      ["Number of phases", 'm'],
                      ["Current loading", 'A'],
                      ["Current Density", 'J'],
                      ["Cu fillfactor", 'kcu'],
                      ["Therm.Loading", 'AJ'],
                      ["torque", 'torque'],
                      ["Force Density", 'fd'],
                      ['Inductance Ld', 'ld'],
                      ['Inductance Lq', 'lq'],
                      ['Stator Resistance', 'r1'],
                      ['Magn. Flux Psim RMS', 'psim'],
                      ["Speed", 'n'],
                      ['Beta-angle',        'beta'],
                      ["Stator-Fe-Losses noload", 'plfe1_0'],
                      ["Rotor-Fe-Losses noload", 'plfe2_0'],
                      ["Magnet-Losses noload", 'plmag_0'],
                      ["Stator-Fe-Losses  ", 'plfe1'],
                      ["Rotor-Fe-Losses   ", 'plfe2'],
                      ["Magnet-Losses", 'plmag'],
                      ["Cu-Losses", 'plcu'],
                      ["Armature", 'lfe'],
                      ["Efficiency", 'eff']]:

                if contentline.find(v[0]) > -1:
                    si = get_si_factor(contentline)
                    rec = l.split()
                    if v[1] in self.machine and isinstance(
                            self.machine[v[1]], list):
                        self.machine[v[1]].append(si*floatnan(rec[-1]))
                    else:
                        self.machine[v[1]] = si*floatnan(rec[-1])
                    break

        if self.machine['beta'] and len(self.machine['beta']) > 1:
            self.machine['beta'] = self.machine['beta'][1:]
        self.machine['n'] = self.machine['n']/60
        self.machine['lfe'] = 1e-3*self.machine['lfe']
        if self.machine['plfe1']:  # calc sum of losses
            plfe1 = self.machine['plfe1']
            plcu = self.machine.get('plcu', [0.0]*len(plfe1))
            if len(plcu) < len(plfe1):
                self.machine['plcu'] = plcu + [plcu[-1]]*(len(plfe1)-len(plcu))
            self.machine['pltotal'] = [sum(pl)
                                       for pl in zip(*[self.machine[k]
                                                       for k in ('plfe1',
                                                                 'plfe2',
                                                                 'plmag',
                                                                 'plcu')])]
            self.machine['plfe'] = [sum(pl)
                                    for pl in zip(*[self.machine[k]
                                                    for k in ('plfe1',
                                                              'plfe2')])]

    def __read_dq_parameter(self, content):
        if content[1].find('Windings') > -1:  # this is the first section

            for l in content[1:]:
                for v in [['Windings Current', 'i1'],
                          ['Angle I vs. Up', 'beta'],
                          ["LD", 'ld'],
                          ["LQ at nom. current", 'lq'],
                          ["Torque TO", 'torque'],
                          ["Force", 'force'],
                          ["Torque constant Kt", 'kt'],
                          ["Magn.Flux no-load", 'psim0'],
                          ["Voltage Up  (RMS)  no-load", 'up0'],
                          ["Magn.Flux load", 'psim'],
                          ["Voltage Up  load", 'up'],
                          ["Speed", 'speed'],
                          ["Number of Poles", 'npoles'],
                          ["Armature length", 'lfe'],
                          ["Airgap diameter", 'dag']]:
                    if l.find(v[0]) > -1:
                        rec = l.split()
                        self.dqPar[v[1]] = [floatnan(rec[-1])]

            # pdb.set_trace()
            for k in ('speed', 'npoles', 'lfe', 'dag', 'up0', 'up', 'psim0'):
                if k in self.dqPar:
                    self.dqPar[k] = self.dqPar[k][0]
            lfe = self.dqPar['lfe']
            self.dqPar['lfe'] = 1e-3*self.dqPar['lfe']
            for k in ('ld', 'lq', 'psim', 'torque', 'force'):
                if k in self.dqPar:
                    self.dqPar[k][0] = lfe*self.dqPar[k][0]
            if 'torque' in self.dqPar:
                self.dqPar['speed'] = self.dqPar['speed']/60
            if 'dag' in self.dqPar:
                self.dqPar['dag'] = 1e-3*self.dqPar['dag']
            if 'npoles' in self.dqPar:
                self.dqPar['npoles'] = int(self.dqPar['npoles'])
            self.dqPar['i1'] = [self.dqPar['i1'][0]/np.sqrt(2)]
            beta = np.pi*self.dqPar['beta'][0]/180
            iq = np.cos(beta)*self.dqPar['i1'][0]
            id = np.sin(beta)*self.dqPar['i1'][0]
            try:
                w1 = np.pi*self.dqPar['speed']*self.dqPar['npoles']
                r1 = self.machine.get('r1', 0.0)
                uq, ud = (r1*iq + self.dqPar['up'] + id*w1*self.dqPar['ld'][0],
                          r1*id - iq*w1*self.dqPar['lq'][0])
                self.dqPar['u1'] = [np.sqrt(uq**2 + ud**2)]
                self.dqPar['gamma'] = [-np.arctan2(ud, uq)*180/np.pi]
                self.dqPar['psim0'] = lfe*self.dqPar['psim0']
                self.dqPar['phi'] = [self.dqPar['beta'][0] +
                                     self.dqPar['gamma'][0]]
                self.dqPar['cosphi'] = [np.cos(np.pi*phi/180)
                                        for phi in self.dqPar['phi']]
                self.dqPar['i1'].insert(0, 0)
                self.dqPar['u1'].insert(0, self.dqPar.get('up0', 0))
            except KeyError:
                pass

            # if next section is absent
            try:
                self.dqPar['psid'] = [self.dqPar['psim'][0]]
                self.dqPar['psiq'] = [self.dqPar['lq'][0] *
                                      self.dqPar['i1'][-1]]
            except KeyError:
                pass
            return  # end of first section

        # second DQ-Parameter section
        for k in ('i1', 'beta', 'ld', 'lq', 'psim', 'up',
                  'psid', 'psiq', 'torque', 'torque_fe', 'torque_sim',
                  'p2', 'u1_fe', 'u1_sim', 'gamma', 'phi', 'Lho', 'Lh2'):
            self.dqPar[k] = []
        lfe = 1e3*self.dqPar['lfe']

        keys = []
        for l in content:
            rec = self.__findNums(l)
            if len(rec) == 8:
                for k, r in zip(*[keys, rec]):
                    self.dqPar[k].append(floatnan(r))
                self.dqPar['p2'].append(self.dqPar['torque'][-1] *
                                        lfe*2*np.pi*self.dqPar['speed'])
                for k in ('ld', 'lq', 'psim', 'psid', 'up', 'psiq', 'torque'):
                    if self.dqPar[k]:
                        self.dqPar[k][-1] = lfe * self.dqPar[k][-1]
            elif len(rec) == 7:
                self.dqPar['torque_fe'].append(floatnan(rec[2]))
                self.dqPar['torque_sim'].append(floatnan(rec[3]))
                self.dqPar['u1_fe'].append(floatnan(rec[4]))
                self.dqPar['u1_sim'].append(floatnan(rec[5]))
                self.dqPar['gamma'].append(floatnan(rec[6]))
                self.dqPar['phi'].append(floatnan(rec[1]) +
                                         self.dqPar['gamma'][-1])
            elif len(rec) == 5:  # self and mutual inductances
                self.dqPar['Lho'].append(lfe*floatnan(rec[2]))
                self.dqPar['Lh2'].append(lfe*floatnan(rec[3]))
            else:
                headers = l.split()
                try:  # must distinguish ee and pm
                    if headers[5] == 'Voltage':
                        keys = ('i1', 'beta', 'ld', 'lq', 'up',
                                'psid', 'psiq', 'torque')
                    if headers[5] == 'Psi_magn':
                        keys = ('i1', 'beta', 'ld', 'lq', 'psim',
                                'psid', 'psiq', 'torque')
                except:
                    pass
        if self.dqPar['psim']:
            self.dqPar.pop('up', None)
        else:
            self.dqPar.pop('psim', None)
        try:
            w1 = np.pi*self.dqPar['speed']*self.dqPar['npoles']
            r1 = self.machine.get('r1', 0.0)
            beta = np.array(self.dqPar['beta'])/180*np.pi
            iq = np.cos(beta)*self.dqPar['i1']
            id = np.sin(beta)*self.dqPar['i1']
            up = w1*np.array(self.dqPar['psim'])
            ld = np.array(self.dqPar['ld'])
            lq = np.array(self.dqPar['lq'])
            uq = r1*iq + up + id*w1*ld
            ud = r1*id - iq*w1*lq
            self.dqPar['up'] = up.tolist()
            self.dqPar['u1'] = np.sqrt(uq**2 + ud**2).tolist()
            self.dqPar['gamma'] = (-np.arctan2(ud, uq)*180/np.pi).tolist()
            self.dqPar['phi'] = (beta/np.pi*180 + self.dqPar['gamma']).tolist()
            self.dqPar['cosphi'] = [np.cos(np.pi*phi/180)
                                    for phi in self.dqPar['phi']]
            self.dqPar['i1'].insert(0, 0)
            self.dqPar['u1'].insert(0, self.dqPar.get('up0', 0))
        except:
            pass

    def __read_weights(self, content):
        #              Stator-Iron      - Conductors      - Magnets
        #                105.408	     22.542	      0.000
        #              Rotor-Iron       - Conductors       - Magnets
        #                 45.041	      0.000	     17.556
        if self.weights:
            return
        scale = 1  # assume kg unit
        if content[0].split()[-1] == '[gr]':
            scale = 1e-3
        for line in content[2:]:
            rec = line.split()
            if rec[0] != 'Stator-Iron' and rec[0] != 'Rotor-Iron':
                self.weights.append([scale*floatnan(x) for x in rec])

    def __read_areas(self, content):
        #  Area [mm**2]: Stator-Iron           - slots         - Magnets
        #                 16585.2	      2383.1	         0.0
        #  Area [mm**2]: Rotor-Iron            - slots         - Magnets
        #                  6372.7	         0.0	      1930.2
        for line in content:
            rec = line.split()
            if rec and rec[0] != 'Area':
                self.areas.append([floatnan(x) for x in rec])

    def __read_inertia(self, content):
        #  Inertia: GD**2/4 [gr m**2/mm]
        #  Stator :	        0.230195
        #  Rotor  :	        0.011774
        self.inertia = []
        pat = re.compile(r'\[(\w\w) .+\]')
        unit = pat.findall(content[0])
        f = 1
        if unit and unit[-1] == 'gr':
            f = 1e-3
        for line in content[1:]:
            x = line.split()
            if x:
                self.inertia.append(f*floatnan(x[-1]))

    def __read_losses(self, content):
        losses = {}
        i = 0
        # find results for angle:
        while True:
            try:
                l = content[i]
            except IndexError:
                break

            if l.startswith('No'):
                i += 1
                continue

            if l.startswith('Results for Angle I-Up'):
                losses['beta'] = floatnan(l.split(':')[-1])
                losses['current'] = floatnan(
                    content[i+1].split(':')[-1])/np.sqrt(2)
                for k in ('staza', 'stajo', 'rotfe',
                          'magnetJ', 'magnetB', 'total'):
                    losses[k] = 0.0
                i += 3
                continue

            if l.startswith('Cu-'):
                # either Cu-losses-Stator or Cu-Losses-Rotor
                reg = l.split()[0].split('-')[2].lower()
                rec = self.__findNums(content[i+1])
                if len(rec) > 0:
                    if reg == 'stator':
                        losses['plcu1'] = floatnan(rec[0])
                        losses['winding'] = floatnan(rec[0])
                        if len(rec) > 2:
                            losses['r1'] = floatnan(rec[2])
                    else:
                        losses['plcu2'] = floatnan(rec[0])
                        if len(rec) > 2:
                            losses['r2'] = floatnan(rec[2])
                    losses['total'] += floatnan(rec[0])
                i += 2
                continue

            if l.startswith('StZa') or l.startswith('RoZa'):
                rec = self.__findNums(content[i+2])
                if len(rec) == 2:
                    losses['stajo'] = floatnan(rec[1])
                    losses['staza'] = floatnan(rec[0])
                    losses['total'] += losses['staza']+losses['stajo']
                i += 3
                continue

            if l.startswith('StJo') or l.startswith('RoJo') or \
               _statloss.search(l):
                rec = self.__findNums(content[i+2])
                if len(rec) == 2:
                    losses['stajo'] = floatnan(rec[0])
                    losses['staza'] = floatnan(rec[1])
                    losses['total'] += losses['staza']+losses['stajo']
                elif len(rec) == 1:
                    t = l.split(':')[-1].strip()
                    if t == 'Iron':
                        losses['staza'] = floatnan(rec[0])
                    else:
                        losses['stajo'] += floatnan(rec[0])
                    losses['total'] += losses['staza']+losses['stajo']
                i += 3
                continue

            if _rotloss.search(l):
                rec = self.__findNums(content[i+2])
                if len(rec) == 1:
                    rotfe = floatnan(rec[0])
                    losses['rotfe'] += rotfe
                    losses['total'] += rotfe
                i += 3
                continue

            if l.find('Fe-Losses-Rotor') > -1:
                rec = self.__findNums(content[i+3])
                if len(rec) == 2:
                    if (content[i+1].find('Iron') > -1 and content[i+1].find('StJo') > 0 or
                        content[i+1].split() == ['RoZa', 'RoJo']):
                        self.external_rotor = True
                        # TODO: there might be better places to check this
                        losses['staza'] = floatnan(rec[0])
                        losses['stajo'] = floatnan(rec[1])
                        losses['total'] += losses['staza']+losses['stajo']
                    else:
                        losses['rotfe'] = floatnan(rec[1])
                        losses['total'] += losses['rotfe']
                i += 4
                continue

            if l.find('Fe-Losses-Stator') > -1:
                rec = self.__findNums(content[i+3])
                if len(rec) == 2:
                    if content[i+1].split() == ['StJo', 'StZa']:
                        losses['stajo'] = floatnan(rec[0])
                        losses['staza'] = floatnan(rec[1])
                        losses['total'] += losses['staza']+losses['stajo']

                    if content[i+1].split() == ['rotf', '----']:
                        losses['rotfe'] = sum([floatnan(x) for x in rec])
                        losses['total'] += losses['rotfe']
                i += 4
                continue

            if l.find('Magnet-Losses') > -1:
                rec = self.__findNums(content[i+1])
                if len(rec) == 1:
                    losses['magnetJ'] = float(rec[0])
                    # losses['magnetB'] = float(Nan)
                if len(rec) == 2:
                    losses['magnetJ'] = float(rec[0])
                    losses['magnetB'] = float(rec[1])
                losses['total'] += losses['magnetJ']
                i += 2
                continue
            i += 1
        if 'total' in losses:
            losses['totalfe'] = sum([losses[k] for k in ('staza',
                                                         'stajo',
                                                         'rotfe')])
            self.losses.append(losses)

    def __read_hysteresis_eddy_current_losses(self, content):
        losses = dict(fft=dict(), stator=dict(), rotor=dict())
        cols = ('order_mech', 'order_el',
                'freq', 'hyst', 'eddy')
        part = 'fft'  # either stator or rotor or fft
        k = ''
        for i, l in enumerate(content):
            if l.startswith('*************'):
                for part in losses:
                    self.losses[-1][part] = dict()
                    for k in losses[part]:
                        for x in losses[part][k]:
                            x[0] = int(x[0])
                            if len(losses[part][k][0]) == 4:
                                cols = ('order_el', 'freq', 'hyst', 'eddy')
                            if losses[part][k]:
                                self.losses[-1][part][k] = {
                                    k1: l
                                    for k1, l in zip(cols,
                                                     zip(*losses[part][k]))}
                        if not losses[part][k] and cols:
                            self.losses[-1][part][k] = {
                                c: () for c in cols}
                self.__read_losses(content[i+1:])
                break

            if l.find('StJo') > -1 or \
               l.find('RoJo') > -1:
                k = 'stajo'
            elif l.find('StZa') > -1 or \
                    l.find('StatorIron') > -1 or \
                    l.find('RoZa') > -1:
                k = 'staza'
            elif l.find('Iron') > -1 and l.find('Stator') > -1:
                k = 'staza'
            elif l.find('Iron') > -1:
                if self.external_rotor:
                    k = 'staza'
                else:
                    k = 'rotor'
            elif l.find(': Roto') > -1:
                k = 'rotor'
            elif l.find(': Ring') > -1:
                k = 'ring'
            elif l.find('Stat') > -1 and l.find('Stator: ') < 0:
                k = 'stajo'
            elif l.find('Stator: ') > -1:
                part = 'stator'
                k = l.split(':')[-1].strip()
            elif l.find('Rotor: ') > -1:
                part = 'rotor'
                k = l.split(':')[-1].strip()
            else:
                if k and k not in losses[part]:
                    losses[part][k] = []
                try:
                    rec = self.__findNums(l)
                    if 'nan' in rec:
                        continue
                    if len(rec) == 4:
                        losses[part][k].append([floatnan(x) for x in rec])
                    elif len(rec) == 5:  # FEMAG Rel 8.3 with el/mech order
                        losses[part][k].append([floatnan(x)
                                                for i, x in enumerate(rec)])
                    else:
                        part = 'fft'

                except:
                    pass

    def get(self, name, r=None):
        """return value of key name
        name can be a list such as ['torque[1]', 'ripple']
        or a string: 'dqPar'
        """
        try:
            if isinstance(name, str):  # ignore r
                lname, indx = splitindex(name)
                if indx is None:
                    return self.__getattr__(name)
                return self.__getattr__(lname).__getitem__(indx)

            if len(name) > 1:
                lname, indx = splitindex(name[0])
                if r:
                    if indx is None:
                        return self.get(name[1:],
                                        getattr(r, lname))
                    return self.get(name[1:],
                                    getattr(r, lname).__getitem__(indx))

                if indx is None:
                    return self.get(name[1:],
                                    getattr(self, lname))
                return self.get(name[1:],
                                getattr(self, lname).__getitem__(indx))

            lname, indx = splitindex(name[0])
            if r:
                if indx is None:
                    return r.get(name[0])
                return r.get(lname).__getitem__(indx)

            if indx is None:
                return self.__getattr__(name[0])
            return self.__getattr__(lname).__getitem__(indx)
        except (KeyError, IndexError, AttributeError):
            return None

    def __getattr__(self, k):
        return self.__dict__[k]

    def items(self):
        return [(k, self.get(k)) for k in ('version',
                                           'type',
                                           'filename',
                                           'date',
                                           'areas',
                                           'inertia',
                                           'current_angles',
                                           'torque',
                                           'torque_fft',
                                           'psidq',
                                           'psidq_ldq',
                                           'machine',
                                           'lossPar',
                                           'weights',
                                           'weight',
                                           'flux',
                                           'flux_fft',
                                           'wdgfactors',
                                           'airgapInduction',
                                           'airgap',
                                           'magnet',
                                           'scData',
                                           'dqPar',
                                           'ldq',
                                           'losses',
                                           'demag',
                                           'linearForce',
                                           'linearForce_fft',
                                           'powerSituation',
                                           'characteristics') if self.get(k)]

    def __str__(self):
        "return string format of this object"
        if self.type:
            return "\n".join([
                'FEMAG {}: {}'.format(self.version, self.type),
                'File: {}  {}'.format(self.filename, self.date)] +
                ['{}: {}'.format(k, v)
                 for k, v in self.items()])
        return "{}"

    def __repr__(self):
        "representation of this object"
        return self.__str__()


def read(filename):
    """Read BCH/BATCH results from file *filename*."""
    import io
    bchresults = Reader()
    with io.open(filename, encoding='latin1', errors='ignore') as f:
        bchresults.read(f.readlines())
    return bchresults


if __name__ == "__main__":
    import json
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    b = read(filename)
    json.dump({k: v for k, v in b.items()}, sys.stdout)
    # print(b)
